"""
Quality
-------
"""

from typing import Collection
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

import numpy as np
import scipy.sparse as sp  # type: ignore
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.utilities as ut

__all__ = [
    "compute_stdev_logs",
    "compute_projected_folds",
    "compute_similar_query_metacells",
    "compute_outliers_matches",
    "compute_deviant_folds",
    "compute_inner_folds",
    "compute_type_genes_normalized_variances",
    "compute_outliers_fold_factors",
    "count_significant_inner_folds",
]


T = TypeVar("T")


@ut.logged()
@ut.timed_call()
def compute_stdev_logs(
    what: Union[str, ut.Matrix] = "__x__",
    *,
    min_gene_total: int = pr.quality_min_gene_total,
    adata: AnnData,
    gdata: AnnData,
    group: Union[str, ut.Vector] = "metacell",
) -> None:
    """
    Compute the standard deviation of the log (base 2) of the fraction of each gene in the cells of the metacell.

    Ideally, the standard deviation should be ~1/3rd of the ``deviants_min_gene_fold_factor`` (which is ``3`` by
    default), indicating that (all)most cells are within that maximal fold factor. In practice we may see higher values.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where ``what`` is a
    per-variable-per-observation (UMIs) matrix or the name of a per-variable-per-observation annotation containing such
    a matrix.

    In addition, ``gdata`` is assumed to have one (fraction) observation for each metacell, a ``total_umis`` per
    metacell, and use the same genes as ``adata``.

    **Returns**

    Sets the following in ``gdata``:

    Per-Variable Per-Observation (Gene-Cell) Annotations

        ``inner_stdev_log``
            For each gene and metacell, the normalized variance (variance over mean) of the gene in the metacell,
            if it has a sufficient number of UMIs to make this meaningful (otherwise, is 0).

    **Computation Parameters**

    For each metacell:

    1. Compute the log (base 2) of the fractions of the UMIs of each gene in each cell, regularized by 1 UMI.

    2. Compute the standard deviation of these logs for each gene across all cells of each metacell.
    """
    assert list(adata.var_names) == list(gdata.var_names)

    metacell_of_cells = ut.get_o_numpy(adata, group, formatter=ut.groups_description)
    assert gdata.n_obs == np.max(metacell_of_cells) + 1

    umis_per_gene_per_cell = ut.get_vo_proper(adata, what, layout="row_major")

    @ut.timed_call()
    def _single_metacell_stdev_log(metacell_index: int) -> Tuple[ut.NumpyVector, ut.NumpyVector]:
        return _compute_metacell_stdev_log(
            metacell_index=metacell_index,
            metacell_of_cells=metacell_of_cells,
            umis_per_gene_per_cell=umis_per_gene_per_cell,
            min_gene_total=min_gene_total,
        )

    results = list(ut.parallel_map(_single_metacell_stdev_log, gdata.n_obs))
    data = np.concatenate([metacell_fold_factors for _metacell_gene_indices, metacell_fold_factors in results])
    indices = np.concatenate([metacell_gene_indices for metacell_gene_indices, _metacell_fold_factors in results])
    indptr = np.array(
        [0] + [len(metacell_gene_indices) for metacell_gene_indices, _metacell_fold_factors in results], dtype="int64"
    )
    np.cumsum(indptr, out=indptr)

    assert data.dtype == "float32"
    assert indices.dtype == "int32"
    assert indptr.dtype == "int64"

    inner_stdev_log_per_gene_per_metacell = sp.csr_matrix((data, indices, indptr), shape=gdata.shape)
    inner_stdev_log_per_gene_per_metacell.has_sorted_indices = True
    inner_stdev_log_per_gene_per_metacell.has_canonical_format = True
    ut.set_vo_data(gdata, "inner_stdev_log", inner_stdev_log_per_gene_per_metacell)


@ut.logged()
def _compute_metacell_stdev_log(
    *,
    metacell_index: int,
    metacell_of_cells: ut.NumpyVector,
    umis_per_gene_per_cell: ut.ProperMatrix,
    min_gene_total: int,
) -> Tuple[ut.NumpyVector, ut.NumpyVector]:
    ut.log_calc("metacell_index", metacell_index)
    cell_indices = np.where(metacell_of_cells == metacell_index)[0]
    umis_per_gene_per_cell_by_rows = ut.to_numpy_matrix(umis_per_gene_per_cell[cell_indices, :])
    umis_per_gene_per_cell_by_columns = ut.to_layout(umis_per_gene_per_cell_by_rows, layout="column_major")
    umis_per_gene = ut.sum_per(umis_per_gene_per_cell_by_columns, per="column")
    genes_mask = umis_per_gene > min_gene_total
    ut.log_calc("genes_mask", genes_mask)
    genes_indices = np.where(genes_mask)[0].astype("int32")
    if len(genes_indices) == 0:
        return (genes_indices, genes_indices.astype("float32"))

    regularized_umis_per_gene_per_cell = ut.to_layout(
        umis_per_gene_per_cell_by_rows[:, genes_indices] + 1, layout="row_major"
    )
    fraction_per_gene_per_cell = ut.fraction_by(regularized_umis_per_gene_per_cell, by="row")
    log_fraction_per_gene_per_cell = np.log2(fraction_per_gene_per_cell, out=fraction_per_gene_per_cell)  # type: ignore
    log_fraction_per_gene_per_cell = ut.to_layout(log_fraction_per_gene_per_cell, layout="column_major")
    stdev_per_gene = ut.stdev_per(log_fraction_per_gene_per_cell, per="column").astype("float32")
    ut.log_calc("stdev_per_gene", stdev_per_gene, formatter=ut.sizes_description)

    gene_indices = np.where(genes_mask)[0].astype("int32")
    return (gene_indices, stdev_per_gene)


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_projected_folds(
    qdata: AnnData,
    from_query_layer: str = "corrected_fraction",
    to_query_layer: str = "projected_fraction",
    fold_regularization: float = pr.project_fold_regularization,
    min_significant_gene_umis: float = pr.project_min_significant_gene_umis,
) -> None:
    """
    Compute the projected fold factors of genes for each query metacell.

    This computes, for each metacell of the query, the fold factors between the corrected and projected gene fractions
    projection of the metacell onto the atlas (see :py:func:`metacells.tools.project.compute_projection_weights`).

    **Input**

    Annotated query ``qdata``, where the observations are query metacells and the variables are genes, where ``what`` is
    a per-variable-per-observation matrix or the name of a per-variable-per-observation annotation containing such a
    matrix.

    In addition, the ``projected`` UMIs of each query metacells onto the atlas.

    **Returns**

    Sets the following in ``qdata``:

    Per-Variable Per-Observation (Gene-Cell) Annotations
        ``projected_fold``
            For each gene and query metacell, the fold factor of this gene between the query and its projection.

    **Computation Parameters**

    1. For each group (metacell), for each gene, compute the gene's fold factor log2((``from_query_layer`` (default:
       {from_query_layer}) + ``fold_regularization``) / (``to_query_layer`` (default: {to_query_layer}) fractions +
       ``fold_regularization``)), similarly to :py:func:`metacells.tools.project.compute_projection_weights` (the
       default ``fold_regularization`` is {fold_regularization}).

    2. Set the fold factor to zero for every case where the total UMIs of the gene in the query metacell are not at
       least ``min_significant_gene_umis`` (default: {min_significant_gene_umis}).
    """
    assert fold_regularization >= 0

    corrected_fractions = ut.get_vo_proper(qdata, from_query_layer, layout="row_major")
    projected_fractions = ut.get_vo_proper(qdata, to_query_layer, layout="row_major")

    corrected_fractions = ut.to_numpy_matrix(corrected_fractions, copy=True)
    projected_fractions = ut.to_numpy_matrix(projected_fractions, copy=True)

    corrected_fractions += fold_regularization
    projected_fractions += fold_regularization

    dense_folds = np.log2(corrected_fractions) - np.log2(projected_fractions)

    total_umis = ut.to_numpy_matrix(ut.get_vo_proper(qdata, "total_umis", layout="row_major"))
    insignificant_folds_mask = total_umis < min_significant_gene_umis
    ut.log_calc("insignificant folds mask", insignificant_folds_mask)

    dense_folds[insignificant_folds_mask] = 0.0
    sparse_folds = sp.csr_matrix(dense_folds)
    sparse_folds.has_sorted_indices = True
    sparse_folds.has_canonical_format = True
    ut.set_vo_data(qdata, "projected_fold", sparse_folds)


@ut.logged()
@ut.timed_call()
def compute_similar_query_metacells(  # pylint: disable=too-many-statements
    qdata: AnnData,
    max_projection_fold_factor: float = pr.project_max_projection_fold_factor,
    max_projection_noisy_fold_factor: float = pr.project_max_projection_noisy_fold_factor,
    min_fitted_query_marker_genes: float = 0,
    max_misfit_genes: int = pr.project_max_misfit_genes,
    essential_genes_property: Union[None, str, Collection[str]] = None,
    min_essential_genes: Optional[int] = None,
    fitted_genes_mask: Optional[ut.NumpyVector] = None,
) -> None:
    """
    Mark query metacells that are "similar" to their projection on the atlas.

    This does not guarantee the query metacell is "the same as" its projection on the atlas; rather, it means the two
    are "sufficiently similar" that one can be reasonably confident in applying atlas metadata to the query metacell
    based on the projection.

    **Input**

    Annotated query ``qdata``, where the observations are metacells and the variables are genes.

    The data should contain per-observation-per-variable annotations ``projected_fold`` with the significant projection
    folds factors, as computed by :py:func:`compute_projected_folds`. If ``min_essential_significant_genes_fraction``,
    and ``essential_genes_property`` are specified, then the data may contain additional per-observation (gene) mask(s)
    denoting the essential genes.

    If a ``projected_noisy_gene`` mask exists, then the genes in it allow for a higher fold factor than normal genes.

    **Returns**

    Sets the following in ``qdata``:

    Per-Observation (Cell) Annotations

        ``similar``
            A boolean mask indicating the query metacell is similar to its projection in the atlas.

    Per-Variable Per-Observation (Gene-Cell) Annotations
        ``misfit``
            Whether the gene has a too-high fold factor between the query and its projection in the atlas.

    **Computation Parameters**

    1. If ``fitted_genes_mask`` is not ``None``, restrict the analysis to the genes listed in it.

    2. Mark as dissimilar any query metacells which have more than ``max_misfit_genes`` (default:
       {max_misfit_genes}) genes whose projection fold is above ``max_projection_fold_factor``,
       or, for genes in ``projected_noisy_gene``, above an additional ``max_projection_noisy_fold_factor``.

    3. Mark as dissimilar any query metacells which did not fit at least ``min_fitted_query_marker_genes`` of the
       query marker genes.

    4. If ``essential_genes_property`` and ``min_essential_genes`` are specified, the former should be the name(s) of
       boolean per-gene property/ies, and we will mark as dissimilar any query metacells which have at least this number
       of essential genes with a low projection fold factor.
    """
    assert max_projection_fold_factor >= 0
    assert max_projection_noisy_fold_factor >= 0
    assert max_misfit_genes >= 0

    projected_fold_per_gene_per_metacell = ut.get_vo_proper(qdata, "projected_fold", layout="row_major")
    projected_fold_per_gene_per_metacell = np.abs(projected_fold_per_gene_per_metacell)  # type: ignore

    if ut.has_data(qdata, "projected_noisy_gene"):
        max_projection_fold_factor_per_gene = np.full(qdata.n_vars, max_projection_fold_factor, dtype="float32")
        noisy_per_gene = ut.get_v_numpy(qdata, "projected_noisy_gene")
        max_projection_fold_factor_per_gene[noisy_per_gene] += max_projection_noisy_fold_factor
        misfit_per_gene_per_metacell_proper = (
            projected_fold_per_gene_per_metacell > max_projection_fold_factor_per_gene[np.newaxis, :]
        )
    else:
        misfit_per_gene_per_metacell_proper = projected_fold_per_gene_per_metacell > max_projection_fold_factor
    misfit_per_gene_per_metacell = sp.csr_matrix(misfit_per_gene_per_metacell_proper)
    misfit_per_gene_per_metacell.has_sorted_indices = True
    misfit_per_gene_per_metacell.has_canonical_format = True
    ut.set_vo_data(qdata, "misfit", misfit_per_gene_per_metacell)

    if fitted_genes_mask is None:
        misfit_per_fitted_gene_per_metacell = misfit_per_gene_per_metacell
    else:
        misfit_per_fitted_gene_per_metacell = misfit_per_gene_per_metacell[:, fitted_genes_mask]

    misfit_per_metacell = ut.sum_per(misfit_per_fitted_gene_per_metacell, per="row")
    ut.log_calc("misfit_per_metacell", misfit_per_metacell, formatter=ut.sizes_description)

    similar_mask = misfit_per_metacell <= max_misfit_genes
    ut.log_calc("similar_mask", similar_mask)

    query_marker_genes_mask = ut.get_v_numpy(qdata, "marker_gene")
    if fitted_genes_mask is not None:
        fitted_query_marker_genes_mask = query_marker_genes_mask & fitted_genes_mask
        ut.log_calc("fitted_query_marker_genes_mask", fitted_query_marker_genes_mask)
    else:
        fitted_query_marker_genes_mask = query_marker_genes_mask
    fitted_query_marker_genes = np.sum(fitted_query_marker_genes_mask)
    ut.log_calc("fitted_query_marker_genes", fitted_query_marker_genes)
    if fitted_query_marker_genes < min_fitted_query_marker_genes:
        similar_mask[:] = False
        ut.log_calc("similar_mask", similar_mask)

    if essential_genes_property is not None and min_essential_genes is not None:
        essential_genes_mask = np.zeros(qdata.n_vars, dtype="bool")
        if isinstance(essential_genes_property, str):
            essential_genes_property = [essential_genes_property]
        for property_name in essential_genes_property:
            essential_genes_mask |= ut.get_v_numpy(qdata, property_name)

        ut.log_calc("essential_genes_mask", essential_genes_mask)
        essential_genes_count = np.sum(essential_genes_mask)
        if essential_genes_count > 0:
            ut.log_calc("essential_gene_names", qdata.var_names[essential_genes_mask])
            misfit_per_essential_gene_per_metacell = misfit_per_gene_per_metacell[:, essential_genes_mask]
            misfit_essential_per_metacell = ut.sum_per(misfit_per_essential_gene_per_metacell, per="row")
            fitted_essential_per_metacell = essential_genes_count - misfit_essential_per_metacell
        else:
            fitted_essential_per_metacell = np.zeros(len(misfit_per_metacell), dtype="bool")

        ut.log_calc("fitted_essential_per_metacell", fitted_essential_per_metacell, formatter=ut.sizes_description)
        essential_similar_mask = fitted_essential_per_metacell >= min_essential_genes
        ut.log_calc("essential_similar_mask", essential_similar_mask)

        similar_mask &= essential_similar_mask

    ut.set_o_data(qdata, "similar", similar_mask)


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_outliers_matches(
    what: Union[str, ut.Matrix] = "__x__",
    *,
    adata: AnnData,
    gdata: AnnData,
    group: Union[str, ut.Vector] = "metacell",
    most_similar: str = "most_similar",
    value_regularization: float = pr.outliers_fold_regularization,
    reproducible: bool,
) -> None:
    """
    Given an assignment of observations (cells) to groups (metacells), compute for each outlier the "most similar"
    group.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where ``what`` is a
    per-variable-per-observation matrix or the name of a per-variable-per-observation annotation containing such a
    matrix.

    In addition, ``gdata`` is assumed to have one observation for each group, and use the same genes as ``adata``. Note
    that there's no requirement that the ``gdata`` will contain the groups defined in ``adata``. That is, it is possible
    to give query cells data in ``adata`` and atlas metacells in ``gdata`` to find the most similar atlas metacell for
    each outlier query metacell.

    **Returns**

    Sets the following in ``adata``:

    Per-Observation (Cell) Annotations

        ``most_similar`` (default: {most_similar})
            For each observation (cell), the index of the "most similar" group.

    **Computation Parameters**

    1. Compute the log2 of the fraction of each gene in each of the outlier cells and the group metacells using
       the ``value_regularization`` (default: {value_regularization}).

    2. Cross-correlate each of the outlier cells with each of the group metacells, in a ``reproducible`` manner.
    """
    assert list(adata.var_names) == list(gdata.var_names)
    cells_similar_metacell_indices = np.full(adata.n_obs, -1, dtype="int32")

    metacell_per_cell = ut.get_o_numpy(adata, group)
    outliers_mask = metacell_per_cell < 0
    if not np.any(outliers_mask):
        ut.set_o_data(adata, most_similar, cells_similar_metacell_indices)
        return

    odata = ut.slice(adata, obs=outliers_mask)

    outliers_data = ut.get_vo_proper(odata, what, layout="row_major")
    metacells_data = ut.get_vo_proper(gdata, what, layout="row_major")

    outliers_fractions = ut.fraction_by(outliers_data, by="row")
    metacells_fractions = ut.fraction_by(metacells_data, by="row")

    outliers_fractions = ut.to_numpy_matrix(outliers_fractions)
    metacells_fractions = ut.to_numpy_matrix(metacells_fractions)

    outliers_fractions += value_regularization
    metacells_fractions += value_regularization

    outliers_log_fractions = np.log2(outliers_fractions, out=outliers_fractions)
    metacells_log_fractions = np.log2(metacells_fractions, out=metacells_fractions)

    outliers_metacells_correlation = ut.cross_corrcoef_rows(
        outliers_log_fractions, metacells_log_fractions, reproducible=reproducible
    )
    outliers_similar_metacell_indices = np.argmax(outliers_metacells_correlation, axis=1)
    assert len(outliers_similar_metacell_indices) == odata.n_obs

    cells_similar_metacell_indices[outliers_mask] = outliers_similar_metacell_indices
    ut.set_o_data(adata, most_similar, cells_similar_metacell_indices)


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_deviant_folds(
    what: Union[str, ut.Matrix] = "__x__",
    *,
    adata: AnnData,
    gdata: AnnData,
    group: Union[str, ut.Vector] = "metacell",
    most_similar: Union[str, ut.Vector, None] = "most_similar",
    min_gene_total: int = pr.quality_min_gene_total,
) -> None:
    """
    Given an assignment of observations (cells) to groups (metacells) or, if an outlier, to the most similar groups,
    compute for each observation and gene the fold factor relative to its group for the purpose of detecting deviant
    cells.

    Ideally, all grouped cells would have no genes with high enough fold factors to be considered deviants, and all
    outlier cells would. In practice grouped cells might have a (few) such genes to the restriction on the fraction
    of deviants.

    It is important not to read too much into the results for a single cell, but looking at which genes appear for cell
    populations (e.g., cells with specific metadata such as batch identification) might be instructive.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where ``what`` is a
    per-variable-per-observation matrix or the name of a per-variable-per-observation annotation containing such a
    matrix.

    In addition, ``gdata`` is assumed to have one observation for each group, and use the same genes as ``adata``.

    **Returns**

    Sets the following in ``adata``:

    Per-Variable Per-Observation (Gene-Cell) Annotations

        ``deviant_fold``
            The fold factor between the cell's UMIs and the expected number of UMIs for the purpose of computing
            deviant cells.

    **Computation Parameters**

    1. For each cell, compute the expected UMIs for each gene given the fraction of the gene in the metacells associated
       with the cell (the one it is belongs to, or the most similar one for outliers).

    2. If the number of UMIs in the metacell (for grouped cells), or sum of the UMIs of the gene in an outlier cell and
       the metacell, is less than ``min_gene_total`` (default: {min_gene_total}), set the fold factor to 0 as we do not
       have sufficient data to robustly estimate it.
    """
    umis_per_gene_per_cell = ut.get_vo_proper(adata, what, layout="row_major")
    fraction_per_gene_per_metacell = ut.get_vo_proper(gdata, what, layout="row_major")
    umis_per_gene_per_metacell = ut.get_vo_proper(gdata, "total_umis", layout="row_major")

    metacell_per_cell = ut.get_o_numpy(adata, group, formatter=ut.groups_description)
    combined_per_cell = metacell_per_cell.copy()
    outliers_mask: Optional[ut.NumpyVector] = None
    if most_similar is not None:
        most_similar_per_cell = ut.get_o_numpy(adata, most_similar, formatter=ut.groups_description)
        outliers_mask = metacell_per_cell < 0
        assert outliers_mask is not None
        assert np.all(most_similar_per_cell[outliers_mask] >= 0)
        assert np.all(most_similar_per_cell[~outliers_mask] < 0)
        combined_per_cell[outliers_mask] = most_similar_per_cell[outliers_mask]

    @ut.timed_call()
    def _single_cell_deviant_folds(cell_index: int) -> Tuple[ut.NumpyVector, ut.NumpyVector]:
        return _compute_cell_deviant_folds(
            cell_index=cell_index,
            umis_per_gene_per_cell=umis_per_gene_per_cell,
            fraction_per_gene_per_metacell=fraction_per_gene_per_metacell,
            umis_per_gene_per_metacell=umis_per_gene_per_metacell,
            outliers_mask=outliers_mask,
            metacell_per_cell=combined_per_cell,
            min_gene_total=min_gene_total,
        )

    results = list(ut.parallel_map(_single_cell_deviant_folds, adata.n_obs))
    data = np.concatenate([cell_fold_factors for _cell_gene_indices, cell_fold_factors in results])
    indices = np.concatenate([cell_gene_indices for cell_gene_indices, _cell_fold_factors in results])
    indptr = np.array(
        [0] + [len(cell_gene_indices) for cell_gene_indices, _cell_fold_factors in results], dtype="int64"
    )
    np.cumsum(indptr, out=indptr)

    assert data.dtype == "float32"
    assert indices.dtype == "int32"
    assert indptr.dtype == "int64"

    fold_factors_per_gene_per_cell = sp.csr_matrix((data, indices, indptr), shape=adata.shape)
    fold_factors_per_gene_per_cell.has_sorted_indices = True
    fold_factors_per_gene_per_cell.has_canonical_format = True
    ut.set_vo_data(adata, "deviant_fold", fold_factors_per_gene_per_cell)


@ut.logged()
def _compute_cell_deviant_folds(
    *,
    cell_index: int,
    umis_per_gene_per_cell: ut.Matrix,
    fraction_per_gene_per_metacell: ut.Matrix,
    umis_per_gene_per_metacell: ut.Matrix,
    metacell_per_cell: ut.NumpyVector,
    outliers_mask: Optional[ut.NumpyVector],
    min_gene_total: int,
) -> Tuple[ut.NumpyVector, ut.NumpyVector]:
    actual_umis_per_gene = ut.to_numpy_vector(umis_per_gene_per_cell[cell_index, :], copy=True)
    total_umis_of_cell = np.sum(actual_umis_per_gene)

    metacell_index = metacell_per_cell[cell_index]
    metacell_fraction_per_gene = ut.to_numpy_vector(fraction_per_gene_per_metacell[metacell_index, :])
    expected_umis_per_gene = total_umis_of_cell * metacell_fraction_per_gene

    actual_umis_per_gene += 1
    expected_umis_per_gene += 1
    fold_factors = np.log2(actual_umis_per_gene) - np.log2(expected_umis_per_gene)

    metacell_umis_per_gene = ut.to_numpy_vector(umis_per_gene_per_metacell[metacell_index, :])
    if outliers_mask is not None and outliers_mask[cell_index]:
        total_umis_per_gene = actual_umis_per_gene + metacell_umis_per_gene
    else:
        total_umis_per_gene = metacell_umis_per_gene

    gene_indices = np.where(total_umis_per_gene >= min_gene_total)[0].astype("int32")
    ut.log_calc("nnz_fold_genes_count", len(gene_indices))
    return (gene_indices.astype("int32"), fold_factors[gene_indices].astype("float32"))


@ut.logged()
@ut.timed_call()
def compute_inner_folds(
    *,
    adata: AnnData,
    gdata: AnnData,
    group: Union[str, ut.Vector] = "metacell",
) -> None:
    """
    Given ``adata`` with computed ``deviant_fold`` for each gene for each cell, set in ``inner_fold`` in ``gdata``, for
    each gene for each metacell the ``deviant_fold`` with the maximal absolute value.
    """
    group_per_cell = ut.get_o_numpy(adata, group)
    deviant_fold_per_gene_per_cell = ut.get_vo_proper(adata, "deviant_fold")

    @ut.timed_call()
    def _single_metacell_inner_folds(metacell_index: int) -> Tuple[ut.NumpyVector, ut.NumpyVector]:
        return _compute_metacell_inner_folds(
            metacell_index=metacell_index,
            group_per_cell=group_per_cell,
            deviant_fold_per_gene_per_cell=deviant_fold_per_gene_per_cell,
        )

    results = list(ut.parallel_map(_single_metacell_inner_folds, gdata.n_obs))

    data = np.concatenate([metacell_fold_factors for _metacell_gene_indices, metacell_fold_factors in results])
    indices = np.concatenate([metacell_gene_indices for metacell_gene_indices, _metacell_fold_factors in results])
    assert len(data) == len(indices)
    indptr = np.array(
        [0] + [len(metacell_gene_indices) for metacell_gene_indices, _metacell_fold_factors in results], dtype="int64"
    )
    np.cumsum(indptr, out=indptr)

    assert data.dtype == "float32"
    assert indices.dtype == "int32"
    assert indptr.dtype == "int64"

    fold_factors_per_gene_per_metacell = sp.csr_matrix((data, indices, indptr), shape=gdata.shape)
    fold_factors_per_gene_per_metacell.has_sorted_indices = True
    fold_factors_per_gene_per_metacell.has_canonical_format = True
    ut.set_vo_data(gdata, "inner_fold", fold_factors_per_gene_per_metacell)


@ut.logged()
def _compute_metacell_inner_folds(
    *,
    metacell_index: int,
    group_per_cell: ut.NumpyVector,
    deviant_fold_per_gene_per_cell: ut.ProperMatrix,
) -> Tuple[ut.NumpyVector, ut.NumpyVector]:
    cells_mask = group_per_cell == metacell_index
    cells_count = np.sum(cells_mask)
    assert cells_count > 0
    genes_count = deviant_fold_per_gene_per_cell.shape[1]

    tmp_deviant_fold_per_gene_per_cell = ut.to_numpy_matrix(deviant_fold_per_gene_per_cell[cells_mask, :], copy=True)
    tmp_deviant_fold_per_gene_per_cell = ut.to_layout(tmp_deviant_fold_per_gene_per_cell, layout="column_major")
    assert tmp_deviant_fold_per_gene_per_cell.shape == (cells_count, genes_count)

    negative_folds_mask = tmp_deviant_fold_per_gene_per_cell < 0
    tmp_deviant_fold_per_gene_per_cell[negative_folds_mask] *= -1

    max_index_per_gene = np.argmax(tmp_deviant_fold_per_gene_per_cell, axis=0)
    assert len(max_index_per_gene) == genes_count

    tmp_deviant_fold_per_gene_per_cell[negative_folds_mask] *= -1

    max_deviant_fold_per_gene = tmp_deviant_fold_per_gene_per_cell[max_index_per_gene, range(genes_count)]
    assert len(max_deviant_fold_per_gene) == genes_count

    gene_indices = np.where(max_deviant_fold_per_gene != 0.0)[0].astype("int32")
    ut.log_calc("nnz_fold_genes_count", len(gene_indices))
    return (gene_indices.astype("int32"), max_deviant_fold_per_gene[gene_indices].astype("float32"))


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_type_genes_normalized_variances(
    what: Union[str, ut.Matrix] = "__x__",
    *,
    adata: AnnData,
    gdata: AnnData,
    group_property: str = "metacell",
    type_property: str = "type",
    type_gene_normalized_variance_quantile: float = pr.type_gene_normalized_variance_quantile,
) -> None:
    """
    Given metacells annotated data with type annotations, compute for each gene for each type how variable
    it is in the cells of the metacells of that type.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation
    annotation containing such a matrix.

    In addition, ``gdata`` is assumed to have one observation for each group, and use the same
    genes as ``adata``. This should have a type annotation.

    **Returns**

    Sets the following in ``gdata``:

    Per-Variable (gene) Annotations:

        ``normalized_variance_in_<type>``
            For each type, the normalized variance (variance over mean) of the gene in the cells of the
            metacells of this type.

    **Computation Parameters**

    1. For each ``type_property`` (default: {type_property}) of metacell in ``gdata``, for each metacell of this type,
       consider all the cells in ``adata`` whose ``group_property`` (default: {group_property}) is that metacell,
       compute the normalized variance (variance over mean) of each gene's expression level, when normalizing each
       cell's total UMIs to the median in its metacell.

    2. Take the ``type_gene_normalized_variance_quantile`` (default: {type_gene_normalized_variance_quantile}) of the
       normalized variance of each gene across all metacells of each type.
    """
    assert list(adata.var_names) == list(gdata.var_names)

    type_of_metacell = ut.get_o_numpy(gdata, type_property)
    unique_types = np.unique(type_of_metacell)
    ut.log_calc("types_count", len(unique_types))

    metacell_per_cell = ut.get_o_numpy(adata, group_property)
    cells_umis = ut.get_vo_proper(adata, what, layout="row_major")

    @ut.logged()
    @ut.timed_call()
    def _compute_single_type_genes_normalized_variances(type_index: int) -> Tuple[str, ut.NumpyVector]:
        type_name = unique_types[type_index]
        ut.log_calc("type_name", type_name)
        type_metacells_indices = np.where(type_of_metacell == type_name)[0]
        ut.log_calc("type_metacells_count", len(type_metacells_indices))
        type_gene_normalized_variances_per_metacell = np.empty(
            (len(type_metacells_indices), adata.n_vars), dtype="float32"
        )

        for position, metacell_index in enumerate(type_metacells_indices):
            _compute_metacell_genes_normalized_variances(
                metacell_per_cell=metacell_per_cell,
                cells_umis=cells_umis,
                position=position,
                metacell_index=metacell_index,
                type_gene_normalized_variances_per_metacell=type_gene_normalized_variances_per_metacell,
            )

        type_gene_normalized_variances_per_metacell = ut.to_layout(
            type_gene_normalized_variances_per_metacell, layout="column_major"
        )
        return type_name, ut.quantile_per(
            type_gene_normalized_variances_per_metacell, quantile=type_gene_normalized_variance_quantile, per="column"
        )

    for type_name, type_gene_normalized_variances in list(
        ut.parallel_map(_compute_single_type_genes_normalized_variances, len(unique_types))
    ):
        ut.set_v_data(gdata, f"normalized_variance_in_{type_name}", type_gene_normalized_variances)


@ut.logged()
@ut.timed_call()
def _compute_metacell_genes_normalized_variances(
    *,
    metacell_per_cell: ut.NumpyVector,
    cells_umis: ut.Matrix,
    position: int,
    metacell_index: int,
    type_gene_normalized_variances_per_metacell: ut.NumpyVector,
) -> None:
    metacell_cells_mask = metacell_per_cell == metacell_index
    metacell_umis = cells_umis[metacell_cells_mask, :]
    total_umis = ut.sum_per(metacell_umis, per="row")
    median_total_umis = np.median(total_umis)
    ut.log_calc("median_total_umis", median_total_umis)
    metacell_fractions = ut.fraction_by(metacell_umis, by="row")
    metacell_fractions = ut.to_layout(metacell_fractions, layout="column_major")
    metacell_fractions *= median_total_umis
    genes_normalized_variance = ut.normalized_variance_per(metacell_fractions, per="column")
    type_gene_normalized_variances_per_metacell[position, :] = genes_normalized_variance


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_outliers_fold_factors(
    what: Union[str, ut.Matrix] = "__x__",
    *,
    adata: AnnData,
    gdata: AnnData,
    most_similar: Union[str, ut.Vector] = "most_similar",
    min_gene_total: int = pr.quality_min_gene_total,
) -> None:
    """
    Given annotated data which is a slice containing just the outliers, where each has a "most similar" group, compute
    for each observation and gene the fold factor relative to its group.

    All outliers should have at least one (typically several) genes with high fold factors, which are the reason they
    couldn't be merged into their most similar group.

    **Input**

    Annotated ``adata``, where the observations are outlier cells and the variables are genes, where ``what`` is a
    per-variable-per-observation matrix or the name of a per-variable-per-observation annotation containing such a
    matrix.

    In addition, ``gdata`` is assumed to have one observation for each group, and use the same genes as ``adata``.
    It should have a ``marker_gene`` mask.

    **Returns**

    Sets the following in ``adata``:

    Per-Variable Per-Observation (Gene-Cell) Annotations

        ``<most_similar>_fold`` (default: {most_similar}_fold)
            The fold factor between the outlier gene expression and their expression in the most similar group, (unless
            the value is too low to be of interest, in which case it will be zero).

    **Computation Parameters**

    1. For each outlier, compute the expected UMIs for each gene given the fraction of the gene in the metacell
       associated with the outlier by the ``most_similar`` (default: {most_similar}).

    2. If the sum of the UMIs of the gene in cell and the metacell are less than ``min_gene_total`` (default:
       {min_gene_total}), set the fold factor to 0 as we do not have sufficient data to robustly estimate it.
    """
    assert list(adata.var_names) == list(gdata.var_names)

    actual_umis_per_gene_per_outlier = ut.to_numpy_matrix(ut.get_vo_proper(adata, what, layout="row_major"), copy=True)
    total_umis_per_outlier = ut.sum_per(actual_umis_per_gene_per_outlier, per="row")

    most_similar_of_outliers = ut.get_o_numpy(adata, most_similar, formatter=ut.groups_description)
    assert np.min(most_similar_of_outliers) >= 0

    metacells_fractions = ut.to_numpy_matrix(ut.get_vo_proper(gdata, what, layout="row_major"))
    metacells_umis = ut.to_numpy_matrix(ut.get_vo_proper(gdata, "total_umis", layout="row_major"))
    fraction_per_gene_per_most_similar = ut.to_numpy_matrix(metacells_fractions[most_similar_of_outliers, :])
    umis_per_gene_per_most_similar = ut.to_numpy_matrix(metacells_umis[most_similar_of_outliers, :])

    expected_umis_per_gene_per_outlier = fraction_per_gene_per_most_similar * total_umis_per_outlier[:, np.newaxis]
    assert actual_umis_per_gene_per_outlier.shape == expected_umis_per_gene_per_outlier.shape

    actual_umis_per_gene_per_outlier += 1
    expected_umis_per_gene_per_outlier += 1

    fold_factor_per_gene_per_outlier = np.log2(actual_umis_per_gene_per_outlier) - np.log2(
        expected_umis_per_gene_per_outlier
    )

    total_umis_per_gene_per_most_similar = umis_per_gene_per_most_similar[most_similar_of_outliers, :]
    total_umis_per_gene_per_fold = actual_umis_per_gene_per_outlier + total_umis_per_gene_per_most_similar
    fold_factor_per_gene_per_outlier[total_umis_per_gene_per_fold < min_gene_total] = 0.0

    fold_factor_per_gene_per_outlier = sp.csr_matrix(fold_factor_per_gene_per_outlier)
    fold_factor_per_gene_per_outlier.has_sorted_indices = True
    fold_factor_per_gene_per_outlier.has_canonical_format = True
    ut.set_vo_data(adata, f"{most_similar}_fold", fold_factor_per_gene_per_outlier)


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def count_significant_inner_folds(
    adata: AnnData,
    *,
    min_gene_fold_factor: float = pr.significant_gene_fold_factor,
) -> None:
    """
    Given grouped (metacells) data, count for each gene in how many metacells there is at least one cell
    with a fold factor above some threshold.

    **Input**

    Annotated ``adata``, where the observations are metacells and the variables are genes, with an ``inner_fold``
    layer (as computed by ``compute_inner_folds``).

    **Returns**

    Sets the ``significant_inner_folds_count`` annotation, counting for each gene the number of metacells
    where the ``inner_fold`` is at least ``min_gene_fold_factor`` (default: {min_gene_fold_factor}), that is,
    where at least one cell in the metacell has a high fold factor for the gene's expression compared to the
    estimated overall gene expression in the metacell.
    """
    inner_fold_per_gene_per_metacell = ut.get_vo_proper(adata, "inner_fold", layout="column_major")
    significant_per_gene_per_metacell = inner_fold_per_gene_per_metacell >= min_gene_fold_factor  # type: ignore
    significant_per_gene = ut.sum_per(significant_per_gene_per_metacell, per="column")  # type: ignore
    ut.set_v_data(adata, "significant_inner_folds_count", significant_per_gene, formatter=ut.sizes_description)



"""
Progress
--------

This used ``tqdm`` to provide a progress bar while computing the metacells.
"""

import logging
from contextlib import contextmanager
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

from tqdm import tqdm  # type: ignore

import metacells.utilities.logging as utl

__all__ = [
    "progress_bar",
    "progress_bar_slice",
    "did_progress",
    "has_progress_bar",
    "start_progress_bar",
    "end_progress_bar",
    "start_progress_bar_slice",
    "end_progress_bar_slice",
]


# Parameters for the progress bar.
TQDM_KWARGS: Optional[Dict[str, Any]] = None

# Global progress bar object.
PROGRESS_BAR: Optional[tqdm] = None

# The log level when the progress bar was activated.
PROGRESS_LOG_LEVEL = logging.INFO

# The current progress position.
PROGRESS_POSITION = 0

# The current progress bar (slice) final position.
PROGRESS_END = 0

#: The size of the current progress bar (slice)
PROGRESS_SIZE = 0


def has_progress_bar() -> bool:
    """
    Return whether there is an active progress bar.
    """
    return PROGRESS_BAR is not None or TQDM_KWARGS is not None


@contextmanager
def progress_bar(**tqdm_kwargs: Any) -> Any:
    """
    Run some code with a ``tqdm`` progress bar.

    ..note::

        When a progress bar is active, logging is restricted to warnings and errors.
    """
    start_progress_bar(**tqdm_kwargs)
    try:
        result = yield None
        return result
    finally:
        end_progress_bar()


def start_progress_bar(**tqdm_kwargs: Any) -> Any:
    """
    Create a progress bar (but do not show it yet).
    """
    global TQDM_KWARGS
    global PROGRESS_POSITION
    global PROGRESS_END
    global PROGRESS_SIZE
    global PROGRESS_LOG_LEVEL

    assert PROGRESS_BAR is None
    assert TQDM_KWARGS is None

    TQDM_KWARGS = tqdm_kwargs.copy()
    if "bar_format" not in TQDM_KWARGS:
        TQDM_KWARGS["bar_format"] = "{l_bar}{bar}[{elapsed}]"  # NOT F-STRING

    PROGRESS_POSITION = 0
    PROGRESS_END = int(1e15)
    PROGRESS_SIZE = PROGRESS_END
    PROGRESS_LOG_LEVEL = utl.logger().level
    utl.logger().setLevel(logging.WARNING)


def end_progress_bar() -> None:
    """
    End an active progress bar.
    """
    global PROGRESS_BAR
    if PROGRESS_BAR is not None:
        PROGRESS_BAR.close()
        utl.logger().setLevel(PROGRESS_LOG_LEVEL)
        PROGRESS_BAR = None


@contextmanager
def progress_bar_slice(fraction: Optional[float]) -> Any:
    """
    Run some code which will use a slice of the current progress bar.

    This can be nested to split the overall progress bar into smaller and smaller parts to represent a tree of
    computations.

    If ``fraction`` is None, or there is no active progress bar, simply runs the code.
    """
    _show_progress_bar()
    if fraction is None:
        result = yield None
        return result

    old_state = start_progress_bar_slice(fraction)
    try:
        result = yield None
        return result
    finally:
        end_progress_bar_slice(old_state)


def start_progress_bar_slice(fraction: float) -> Tuple[int, int]:
    """
    Start a nested slice of the overall progress bar.

    Returned the captured state that needs to be passed to ``end_progress_bar_slice``.
    """
    assert PROGRESS_BAR is not None
    assert TQDM_KWARGS is None
    assert 0.0 < fraction <= 1.0, f"invalid progress fraction: {fraction} (1 + {fraction - 1})"

    global PROGRESS_END
    global PROGRESS_SIZE

    old_progress_size = PROGRESS_SIZE
    old_progress_end = PROGRESS_END

    PROGRESS_SIZE = int(PROGRESS_SIZE * fraction)
    PROGRESS_END = PROGRESS_POSITION + PROGRESS_SIZE

    assert (
        0 < PROGRESS_SIZE <= old_progress_size
    ), f"invalid progress size: {PROGRESS_SIZE} ({old_progress_size} + {PROGRESS_SIZE - old_progress_size})"

    assert (
        PROGRESS_END <= old_progress_end
    ), f"invalid progress end: {PROGRESS_END} ({old_progress_end} + {PROGRESS_END - old_progress_end})"

    return old_progress_size, old_progress_end


def end_progress_bar_slice(old_state: Tuple[int, int]) -> None:
    """
    End a nested slice of the overall progress bar.

    This moves the progress bar position to the end of the slice regardless of reported progress within it.
    """
    old_progress_size, old_progress_end = old_state

    global PROGRESS_POSITION
    global PROGRESS_END
    global PROGRESS_SIZE

    assert PROGRESS_BAR is not None
    assert TQDM_KWARGS is None
    assert PROGRESS_POSITION <= PROGRESS_END

    remaining = PROGRESS_END - PROGRESS_POSITION
    if remaining > 0:
        PROGRESS_BAR.update(remaining)
    PROGRESS_POSITION = PROGRESS_END
    PROGRESS_SIZE = old_progress_size
    PROGRESS_END = old_progress_end


def did_progress(fraction: float) -> Any:
    """
    Report progress of some fraction of the current (slice of) progress bar.
    """
    _show_progress_bar()
    assert 0.0 < fraction <= 1.0, f"invalid progress fraction: {fraction} (1 + {fraction - 1})"
    assert PROGRESS_BAR is not None
    assert TQDM_KWARGS is None

    global PROGRESS_POSITION

    step = int(PROGRESS_SIZE * fraction)
    step = min(step, PROGRESS_END - PROGRESS_POSITION)

    if step > 0:
        PROGRESS_BAR.update(step)
        PROGRESS_POSITION += step


def _show_progress_bar() -> None:
    global PROGRESS_BAR
    global TQDM_KWARGS
    if PROGRESS_BAR is None and TQDM_KWARGS is not None:
        PROGRESS_BAR = tqdm(total=PROGRESS_SIZE, **TQDM_KWARGS)  # pylint: disable=not-a-mapping
        TQDM_KWARGS = None



"""
Process a ``timing.csv`` file.
"""

import csv
import os
import sys
from argparse import ArgumentParser
from textwrap import dedent
from typing import Dict
from typing import List
from typing import Optional


def main() -> None:
    """
    Process a ``timing.csv`` file.
    """
    parser = ArgumentParser(description="Process a timing.csv file")
    subparsers = parser.add_subparsers(dest="command", metavar="COMMAND", title="commands", help="One of:")

    subparsers.required = True

    combine_parser = subparsers.add_parser(
        "combine",
        help="Combine parallel timing files",
        description=dedent(
            """
        Read a main timings.csv file and all the parallel timing.<map>.<process>.csv files, and
        combine them to a single timing CSV file written to the standard output. In this case, the
        input main timing CSV file name is required (to allow locating the associated parallel
        timing CSV file names).

        The data in the combined file normalizes the data from the parallel map loops such that the
        results show the average elapsed time per process within such loops, but still the total CPU
        time.

        The combined output is in no particular order. Pass it through the sum or flame command
        to obtain a human-readable format.
    """
        ),
    )

    combine_parser.add_argument(
        "-o",
        "--output",
        metavar="OUTPUT",
        help="The optional path to write the output to " "(otherwise, writes to the standard output).",
    )

    combine_parser.add_argument("input", metavar="FILE", help="The path to the timing CSV file.")

    sum_parser = subparsers.add_parser(
        "sum",
        help="Sum the total time for each step",
        description=dedent(
            """
        Read a timings.csv file from the input and write a sum file with one line per
        step containing the sum of the data and the number of invocations to the output.

        The data in the sum file is presented in seconds and billions of instructions, to make it
        easier to follow.

        The output is sorted in descending elapsed time order.

        You can pipe the output through "column -t -s," to make it more legible.
    """
        ),
    )

    sum_parser.add_argument(
        "-o",
        "--output",
        metavar="OUTPUT",
        help="The optional path to write the output to " "(otherwise, writes to the standard output).",
    )

    sum_parser.add_argument(
        "input",
        metavar="FILE",
        nargs="?",
        help="The optional path to the timing CSV file " "(otherwise, reads from the standard input).",
    )

    flame_parser = subparsers.add_parser(
        "flame",
        help="Reformat the data for visualization in flamegraph",
        description=dedent(
            """
        Read a timings.csv file from the input and write a flamegraph file with the
        chosen data (by default, elapsed time). This can be viewed by any flamegraph
        visualization tool such as flameview.
    """
        ),
    )

    flame_parser.add_argument(
        "-f",
        "--focus",
        default="elapsed",
        choices=["elapsed", "cpu", "invocations"],
        help="Which field to focus the flamegraph on.",
    )

    flame_parser.add_argument(
        "-s",
        "--seconds",
        action="store_true",
        help="Output the data in float seconds instead of integer "
        "nanoseconds. Note that most flamegraph viewers can "
        "only handle integer data (flameview can handle float).",
    )

    flame_parser.add_argument(
        "-o",
        "--output",
        metavar="OUTPUT",
        help="The optional path to write the output to " "(otherwise, writes to the standard output).",
    )

    flame_parser.add_argument(
        "input",
        metavar="FILE",
        nargs="?",
        help="The optional path to the timing CSV file " "(otherwise, reads from the standard input).",
    )

    args = parser.parse_args()

    if args.input is not None:
        assert args.input.endswith(".csv")

    if args.command == "combine":
        assert args.input is not None
        _combine_main(args.input, args.output)

    elif args.command == "sum":
        _sum_main(args.input, args.output)

    elif args.command == "flame":
        _flame_main(args.input, args.output, args.focus, args.seconds)


def _combine_main(input_path: str, output_path: Optional[str]) -> None:
    map_index = 0

    if output_path is None:
        output_file = sys.stdout
    else:
        output_file = open(output_path, "w", encoding="utf8")  # pylint: disable=consider-using-with

    for line in open(input_path, "r", encoding="utf8").readlines():  # pylint: disable=consider-using-with
        if ";parallel_map," not in line:
            output_file.write(line)
            continue

        map_index += 1
        line = line.strip()
        fields = line.split(",")
        assert fields[5] == "index"
        assert map_index == int(fields[6])
        assert fields[7] == "processes"
        fields[7] = "expected_processes"
        expected_processes_count = int(fields[8])

        actual_processes_count = expected_processes_count
        process_index = 0
        while True:
            process_index += 1
            process_input = f"{input_path[:-4]}.{map_index}.{process_index}.csv"
            if not os.path.exists(process_input):
                actual_processes_count = process_index
                break

        fields.append("actual_processes")
        fields.append(str(actual_processes_count))
        assert fields[1] == "elapsed_ns"
        fields[2] = str(float(fields[2]) / (actual_processes_count + 1))
        line = ",".join(fields)
        output_file.write(line)
        output_file.write("\n")

        for process_index in range(actual_processes_count):
            process_input = f"{input_path[:-4]}.{map_index}.{process_index}.csv"
            with open(process_input, "r", encoding="utf8") as file:
                for process_line in file.readlines():
                    process_fields = process_line.split(",")
                    assert process_fields[1] == "elapsed_ns"
                    process_fields[2] = str(float(process_fields[2]) / (actual_processes_count + 1))
                    process_line = ",".join(process_fields)
                    output_file.write(process_line)


def _sum_main(input_path: Optional[str], output_path: Optional[str]) -> None:
    data_by_name = _collect_data_by_name(input_path, True)

    total_data: List[float] = []
    for name, data in data_by_name.items():
        while len(total_data) < len(data):
            total_data.append(0)
        for index, datum in enumerate(data):
            total_data[index] += datum

    data_by_name["TOTAL"] = total_data

    if output_path is None:
        output_file = sys.stdout
    else:
        output_file = open(output_path, "w", encoding="utf8")  # pylint: disable=consider-using-with

    fields = ["invocations", "elapsed_s", "cpu_s"]
    for name, data in sorted(data_by_name.items(), key=lambda data: data[1][1], reverse=True):
        text = [name]
        for field, value in zip(fields, data):
            text.append(field)
            if field == "invocations":
                text.append(str(value))
            else:
                text.append(str(value / 1_000_000_000))
        output_file.write(",".join(text))
        output_file.write("\n")


def _flame_main(input_path: Optional[str], output_path: Optional[str], focus: str, seconds: bool) -> None:
    data_by_name = _collect_data_by_name(input_path, False)

    if output_path is None:
        output_file = sys.stdout
    else:
        output_file = open(output_path, "w", encoding="utf8")  # pylint: disable=consider-using-with

    for name, data in data_by_name.items():
        if seconds:
            datum = {"invocations": data[0], "elapsed": data[1] / 1_000_000_000, "cpu": data[2] / 1_000_000_000}
        else:
            datum = {"invocations": data[0], "elapsed": int(round(data[1])), "cpu": int(round(data[2]))}
        html = (
            "Elapsed Time: %.2f<br/>"  # pylint: disable=consider-using-f-string
            "CPU Time: %.2f<br/>"
            "Utilization: %.0f%%<br/>"
            "Invocations: %s<br/>"
            % (datum["elapsed"], datum["cpu"], 100 * datum["cpu"] / datum["elapsed"], datum["invocations"])
        )
        output_file.write(f"{name.replace('.', ';')} {datum[focus]} #{html}\n")


def _collect_data_by_name(input_path: Optional[str], split: bool) -> Dict[str, List[float]]:
    data_by_name: Dict[str, List[float]] = {}

    if input_path is None:
        input_file = sys.stdin
    else:
        input_file = open(input_path, "r", encoding="utf8")  # pylint: disable=consider-using-with

    for row in csv.reader(input_file):
        name = row[0]
        if split:
            name = name.split(";")[-1]
        assert row[1] == "elapsed_ns"
        elapsed_ns = float(row[2])
        assert row[3] == "cpu_ns"
        cpu_ns = float(row[4])

        data = data_by_name.get(name)
        if data is None:
            data = [0, 0, 0]
            data_by_name[name] = data
        data[0] += 1
        data[1] += elapsed_ns
        data[2] += cpu_ns

    return data_by_name


if __name__ == "__main__":
    main()



"""
Projection
----------
"""

from math import ceil
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import numpy as np
import scipy.sparse as sp  # type: ignore
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.tools as tl
import metacells.utilities as ut
from metacells import __version__  # pylint: disable=cyclic-import

__all__ = [
    "projection_pipeline",
    "outliers_projection_pipeline",
    "write_projection_weights",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def projection_pipeline(
    what: str = "__x__",
    *,
    adata: AnnData,
    qdata: AnnData,
    only_atlas_marker_genes: bool = pr.only_atlas_marker_genes,
    only_query_marker_genes: bool = pr.only_query_marker_genes,
    ignore_atlas_lateral_genes: bool = pr.ignore_atlas_lateral_genes,
    ignore_query_lateral_genes: bool = pr.ignore_query_lateral_genes,
    consider_atlas_noisy_genes: bool = pr.consider_atlas_noisy_genes,
    consider_query_noisy_genes: bool = pr.consider_query_noisy_genes,
    misfit_min_metacells_fraction: float = pr.misfit_min_metacells_fraction,
    project_log_data: bool = pr.project_log_data,
    project_fold_regularization: float = pr.project_fold_regularization,
    project_candidates_count: int = pr.project_candidates_count,
    project_min_candidates_fraction: float = pr.project_min_candidates_fraction,
    project_min_significant_gene_umis: int = pr.project_min_significant_gene_umis,
    project_min_usage_weight: float = pr.project_min_usage_weight,
    project_filter_ranges: bool = pr.project_filter_ranges,
    project_ignore_range_quantile: float = pr.project_ignore_range_quantile,
    project_ignore_range_min_overlap_fraction: float = pr.project_ignore_range_min_overlap_fraction,
    project_min_query_markers_fraction: float = pr.project_min_query_markers_fraction,
    project_max_consistency_fold_factor: float = pr.project_max_consistency_fold_factor,
    project_max_projection_fold_factor: float = pr.project_max_projection_fold_factor,
    project_max_projection_noisy_fold_factor: float = pr.project_max_projection_noisy_fold_factor,
    project_max_misfit_genes: int = pr.project_max_misfit_genes,
    project_min_essential_genes_fraction: Optional[float] = pr.project_min_essential_genes_fraction,
    atlas_type_property_name: str = "type",
    project_corrections: bool = pr.project_corrections,
    project_min_corrected_gene_correlation: float = pr.project_min_corrected_gene_correlation,
    project_min_corrected_gene_factor: float = pr.project_min_corrected_gene_factor,
    reproducible: bool,
    top_level_parallel: bool = True,
) -> ut.CompressedMatrix:
    """
    Complete pipeline for projecting query metacells onto an atlas of metacells for the ``what`` (default: {what}) data.

    **Input**

    Annotated query ``qdata`` and atlas ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation annotation
    containing such a matrix, containing the fraction of each gene in each cell. The atlas should also contain a
    ``type`` per-observation (metacell) annotation.

    The ``qdata`` may include per-gene masks, ``ignored_gene`` and ``ignored_gene_of_<type>``, which force the code
    below to ignore the marked genes from either the preliminary projection or the following refined type-specific
    projections.

    **Returns**

    A matrix whose rows are query metacells and columns are atlas metacells, where each entry is the weight of the atlas
    metacell in the projection of the query metacells. The sum of weights in each row (that is, for a single query
    metacell) is 1. The weighted sum of the atlas metacells using these weights is the "projected" image of the query
    metacell onto the atlas.

    Variable (Gene) Annotations
        ``atlas_gene``
            A mask of the query genes that also exist in the atlas. We match genes by their name; if projecting query
            data from a different technology, we expect the caller to modify the query gene names to match the atlas
            before projecting it.

        ``atlas_lateral_gene``, ``atlas_noisy_gene``, ``atlas_marker_gene``, ``essential_gene_of_<type>``
            Copied from the atlas to the query (``False`` for non-``atlas_gene``).

        ``projected_noisy_gene``
            The mask of the genes that were considered "noisy" when computing the projection. By default this is the
            union of the noisy atlas and query genes.

        ``correction_factor`` (if ``project_corrections``)
            If projecting a query on an atlas with different technologies (e.g., 10X v3 to 10X v2), an automatically
            computed factor we multiplied the query gene fractions by to compensate for the systematic difference
            between the technologies (1.0 for uncorrected genes and 0.0 for non-``atlas_gene``).

        ``fitted_gene_of_<type>``
            For each type, the genes that were projected well from the query to the atlas for most cells of that
            type; any ``atlas_gene`` outside this mask failed to project well from the query to the atlas for most
            metacells of this type. For non-``atlas_gene`` this is set to ``False``.

        ``misfit_gene_of_<type>``
            For each query metacell type, a boolean mask indicating whether the gene has a strong bias in the query
            metacells of this type compared to the atlas metacells of this type.

        ``ignored_gene_of_<type>``
            For each query metacell type, a boolean mask indicating whether the gene was ignored by the projection (for
            any reason) when computing the projection for metacells of this type.

    Observation (Cell) Annotations
        ``total_atlas_umis``
            The total UMIs of the ``atlas_gene`` in each query metacell. This is used in the analysis as described for
            ``total_umis`` above, that is, to ensure comparing expression levels will ignore cases where the total
            number of UMIs of both compared gene profiles is too low to make a reliable determination. In such cases we
            take the fold factor to be 0.

        ``projected_type``
            For each query metacell, the best atlas ``type`` we can assign to it based on its projection. Note this does
            not indicate that the query metacell is "truly" of this type; to make this determination one needs to look
            at the quality control data below.

        ``projected_secondary_type``
            In some cases, a query metacell may fail to project well to a single region of the atlas, but does project
            well to a combination of two distinct atlas regions. This may be due to the query metacell containing
            doublets, of a mixture of cells which match different atlas regions (e.g. due to sparsity of data in the
            query data set). Either way, if this happens, we place here the type that best describes the secondary
            region the query metacell was projected to; otherwise this would be the empty string. Note that the
            ``weights`` matrix above does not distinguish between the regions.

        ``projected_correlation`` per query metacell
            The correlation between between the ``corrected_fraction`` and the ``projected_fraction`` for only the
            ``fitted_gene`` expression levels of each query metacell. This serves as a very rough estimator for the
            quality of the projection for this query metacell (e.g. can be used to compute R^2 values).

            In general we expect high correlation (more than 0.9 in most metacells) since we restricted the
            ``fitted_gene`` mask only to genes we projected well.

        ``similar`` mask per query metacell
            A conservative determination of whether the query metacell is "similar" to its projection on the atlas. This
            is based on whether the number of ``misfit_gene`` for the query metacell is low enough (by default, up to 3
            genes), and also that at least 75% of the ``essential_gene`` of the query metacell were not ``misfit_gene``.
            Note that this explicitly allows for a ``projected_secondary_type``, that is, a metacell of doublets will be
            "similar" to the atlas, but a metacell of a novel state missing from the atlas will be "dissimilar".

            The final determination of whether to accept the projection is, as always, up to the analyst, based on prior
            biological knowledge, the context of the collection of the query (and atlas) data sets, etc. The analyst
            need not (indeed, *should not*) blindly accept the ``similar`` determination without examining the rest of
            the quality control data listed above.

    Observation-Variable (Cell-Gene) Annotations
        ``corrected_fraction`` per gene per query metacell
            For each ``atlas_gene``, its fraction in each query metacell, out of only the atlas genes. This may be
            further corrected (see below) if projecting between different scRNA-seq technologies (e.g. 10X v2 and 10X
            v3). For non-``atlas_gene`` this is 0.

        ``projected_fraction`` per gene per query metacell
            For each ``atlas_gene``, its fraction in its projection on the atlas. This projection is computed as a
            weighted average of some atlas metacells (see below), which are all sufficiently close to each other (in
            terms of gene expression), so averaging them is reasonable to capture the fact the query metacell may be
            along some position on some gradient that isn't an exact match for any specific atlas metacell. For
            non-``atlas_gene`` this is 0.

        ``fitted`` mask per gene per query metacell
            For each ``atlas_gene`` for each query metacell, whether the gene was expected to be projected well, based
            on the query metacell ``projected_type`` (and the ``projected_secondary_type``, if any). For
            non-``atlas_gene`` this is set to ``False``. This does not guarantee the gene was actually projected well.

        ``misfit``
            For each ``atlas_gene`` for each query metacell, whether the ``corrected_fraction`` of the gene was
            significantly different from the ``projected_fractions`` (that is, whether the gene was not projected well
            for this metacell). For non-``atlas_gene`` this is set to ``False``, to make it easier to identify
            problematic genes.

            This is expected to be rare for ``fitted_gene`` and common for the rest of the ``atlas_gene``. If too many
            ``fitted_gene`` are also ``misfit_gene``, then one should be suspicious whether the query metacell is
            "truly" of the ``projected_type``.

        ``essential``
            Which of the ``atlas_gene`` were also listed in the ``essential_gene_of_<type>`` for the ``projected_type``
            (and also the ``projected_secondary_type``, if any) of each query metacell.

            If an ``essential_gene`` is also a ``misfit_gene``, then one should be very suspicious whether the query
            metacell is "truly" of the ``projected_type``.

        ``projected_fold`` per gene per query metacell
            The fold factor between the ``corrected_fraction`` and the ``projected_fraction`` (0 for
            non-``atlas_gene``). If the absolute value of this is high (3 for 8x ratio) then the gene was not projected
            well for this metacell. This will be 0 for non-``atlas_gene``.

            It is expected this would have low values for most ``fitted_gene`` and high values for the rest of the
            ``atlas_gene``, but specific values will vary from one query metacell to another. This allows the analyst to
            make fine-grained determination about the quality of the projection, and/or identify quantitative
            differences between the query and the atlas (e.g., when studying perturbed systems such as knockouts or
            disease models).

    **Computation Parameters**

    0. Find the subset of genes that exist in both the query and the atlas. All computations will be done on this common
       subset. Normalize the fractions of the fitted gene fractions to sum to 1 in each metacell.

    Compute preliminary projection:

    1. Compute a mask of fitted genes, ignoring any genes included in ``ignored_gene``. If ``only_atlas_marker_genes``
       (default: ``only_atlas_marker_genes``), ignore any non-``marker_gene`` of the atlas. If
       ``only_query_marker_genes`` (default: ``only_query_marker_genes``), ignore any non-``marker_gene`` of the query.
       If ``ignore_atlas_lateral_genes`` (default: {ignore_atlas_lateral_genes}), ignore the ``lateral_gene`` of the
       atlas. If ``ignore_query_lateral_genes`` (default: {ignore_query_lateral_genes}), ignore the ``lateral_gene`` of
       the atlas. Normalize the fractions of the fitted gene fractions to sum to 1 in each metacell.

    2. Invoke :py:func:`metacells.tools.project.compute_projection_weights`
       to project each query metacell onto the atlas, using the ``project_log_data`` (default: {project_log_data}),
       ``project_fold_regularization`` (default: {project_fold_regularization}), ``project_min_significant_gene_umis``
       (default: {project_min_significant_gene_umis}), ``project_max_consistency_fold_factor`` (default:
       {project_max_consistency_fold_factor}), ``project_candidates_count`` (default: {project_candidates_count}),
       ``project_min_usage_weight`` (default: {project_min_usage_weight}), and ``reproducible``.

    3. If ``project_corrections`` (default: {project_corrections}: Correlate the expression levels of each gene between
       the query and projection. If this is at least ``project_min_corrected_gene_correlation`` (default:
       {project_min_corrected_gene_correlation}), compute the ratio between the mean expression of the gene in the
       projection and the query. If this is at most 1/(1+``project_min_corrected_gene_factor``) or at least
       (1+``project_min_corrected_gene_factor``) (default: {project_min_corrected_gene_factor}), then multiply the
       gene's value by this factor so its level would match the atlas. As usual, ignore genes which do not have at least
       ``project_min_significant_gene_umis``. If any genes were corrected, then repeat steps 2-3 (but do these steps no
       more than 3 times).

    4. If ``project_filter_ranges`` (default: {project_filter_ranges}): Compute for each gene its expression range
       (lowest and highest ``project_ignore_range_quantile`` (default: {project_ignore_range_quantile}) in both the
       projected and the corrected values. Compute the overlap between these ranges (shared range divided by the query
       range). If this is less than ``project_ignore_range_min_overlap_fraction`` (default:
       {project_ignore_range_min_overlap_fraction}), then ignore the gene. If any genes were ignored, repeat steps 2-4
       (but do this step no more than 3 times).

    5. Invoke :py:func:`metacells.tools.project.convey_atlas_to_query` to assign a projected type to each of the
       query metacells based on the ``atlas_type_property_name`` (default: {atlas_type_property_name}).

    Then, for each type of query metacells:

    If ``top_level_parallel`` (default: ``top_level_parallel``), do this in parallel. This seems to work better than
    doing this serially and using parallelism within each type. However this is still inefficient compared to using both
    types of parallelism at once, which the code currently can't do without non-trivial coding (this would have been
    trivial in Julia...).

    6. Further reduce the mask of type-specific fitted genes by ignoring any genes in ``ignored_gene_of_<type>``, if
       this annotation exists in the query. Normalize the sum of the fitted gene fractions to 1 in each metacell.

    7. Invoke :py:func:`metacells.tools.project.compute_projection_weights` to project each query metacell of the type
       onto the atlas. Note that even though we only look at query metacells (tentatively) assigned the type, their
       projection on the atlas may use metacells of any type.

    8. Invoke :py:func:`metacells.tools.quality.compute_projected_folds` to compute the significant fold factors
       between the query and its projection.

    9. Identify type-specific misfit genes whose fold factor is above ``project_max_projection_fold_factor``. If
       ``consider_atlas_noisy_genes`` and/or ``consider_query_noisy_genes``, then any gene listed in either is allowed
       an additional ``project_max_projection_noisy_fold_factor``. If any gene has such a high fold factor in at least
       ``misfit_min_metacells_fraction``, remove it from the fitted genes mask and repeat steps 6-9.

    10. Invoke :py:func:`metacells.tools.quality.compute_similar_query_metacells` to verify which query metacells
        ended up being sufficiently similar to their projection, using ``project_max_consistency_fold_factor`` (default:
        {project_max_consistency_fold_factor}), ``project_max_projection_noisy_fold_factor`` (default:
        {project_max_projection_noisy_fold_factor}), ``project_max_misfit_genes`` (default: {project_max_misfit_genes}),
        and if needed, the the ``essential_gene_of_<type>``. Also compute the correlation between the corrected
        and projected gene fractions for each metacell.

    And then:

    11. Invoke :py:func:`metacells.tools.project.convey_atlas_to_query` to assign an updated projected type to each of
        the query metacells based on the ``atlas_type_property_name`` (default: {atlas_type_property_name}). If this
        changed the type assigned to any query metacell, repeat steps 6-11 (but do this step no more than 3 times).

    For each query metacell that ended up being dissimilar, try to project them as a combination of two atlas regions:

    12. Reduce the list of fitted genes for the metacell based on the ``ignored_gene_of_<type>`` for both the primary
        (initially from the above) query metacell type and the secondary query metacell type (initially empty).
        Normalize the sum of the gene fractions in the metacell to 1.

    13. Invoke :py:func:`metacells.tools.project.compute_projection_weights` just for this metacell, allowing the
        projection to use a secondary location in the atlas based on the residuals of the atlas metacells relative to
        the primary query projection.

    14. Invoke :py:func:`metacells.tools.project.convey_atlas_to_query` twice, once for the weights of the primary
        location and once for the weights of the secondary location, to obtain a primary and secondary type for
        the query metacell. If these have changed, repeat steps 13-14 (but do these steps no more than 3 times; note
        will will always do them twice as the 1st run will generate some non-empty secondary type).

    15. Invoke :py:func:`metacells.tools.quality.compute_projected_folds` and
        :py:func:`metacells.tools.quality.compute_similar_query_metacells` to update the projection and evaluation of
        the query metacell. If it is now similar, then use the results for the metacell; otherwise, keep the original
        results as they were at the end of step 10.
    """
    assert project_min_corrected_gene_factor >= 0
    use_essential_genes = project_min_essential_genes_fraction is not None and _has_any_essential_genes(adata)

    ut.set_m_data(qdata, "project_max_projection_fold_factor", project_max_projection_fold_factor)
    ut.set_m_data(qdata, "project_max_projection_noisy_fold_factor", project_max_projection_noisy_fold_factor)
    ut.set_m_data(qdata, "project_max_misfit_genes", project_max_misfit_genes)

    atlas_common_gene_indices, query_common_gene_indices = _common_gene_indices(adata, qdata)

    type_names = _initialize_atlas_data_in_query(
        adata,
        qdata,
        atlas_type_property_name=atlas_type_property_name,
        atlas_common_gene_indices=atlas_common_gene_indices,
        query_common_gene_indices=query_common_gene_indices,
        consider_atlas_noisy_genes=consider_atlas_noisy_genes,
        consider_query_noisy_genes=consider_query_noisy_genes,
        use_essential_genes=use_essential_genes,
    )

    common_adata = _initialize_common_adata(adata, what, atlas_common_gene_indices)
    common_qdata = _initialize_common_qdata(qdata, what, query_common_gene_indices)

    query_marker_genes_mask = ut.get_v_numpy(common_qdata, "marker_gene")
    min_fitted_query_marker_genes = np.sum(query_marker_genes_mask) * project_min_query_markers_fraction
    ut.log_calc("min_fitted_query_marker_genes", min_fitted_query_marker_genes)

    min_essential_genes_of_type = _min_essential_genes_of_type(
        adata=adata,
        common_adata=common_adata,
        type_names=type_names,
        project_min_essential_genes_fraction=project_min_essential_genes_fraction,
    )

    preliminary_fitted_genes_mask = _preliminary_fitted_genes_mask(
        common_adata=common_adata,
        common_qdata=common_qdata,
        only_atlas_marker_genes=only_atlas_marker_genes,
        only_query_marker_genes=only_query_marker_genes,
        ignore_atlas_lateral_genes=ignore_atlas_lateral_genes,
        ignore_query_lateral_genes=ignore_query_lateral_genes,
        min_significant_gene_umis=project_min_significant_gene_umis,
    )

    _compute_preliminary_projection(
        common_adata=common_adata,
        common_qdata=common_qdata,
        preliminary_fitted_genes_mask=preliminary_fitted_genes_mask,
        project_log_data=project_log_data,
        project_fold_regularization=project_fold_regularization,
        project_min_significant_gene_umis=project_min_significant_gene_umis,
        project_filter_ranges=project_filter_ranges,
        project_ignore_range_quantile=project_ignore_range_quantile,
        project_ignore_range_min_overlap_fraction=project_ignore_range_min_overlap_fraction,
        project_max_consistency_fold_factor=project_max_consistency_fold_factor,
        project_candidates_count=project_candidates_count,
        project_min_usage_weight=project_min_usage_weight,
        project_corrections=project_corrections,
        project_min_corrected_gene_correlation=project_min_corrected_gene_correlation,
        project_min_corrected_gene_factor=project_min_corrected_gene_factor,
        atlas_type_property_name=atlas_type_property_name,
        reproducible=reproducible,
    )

    weights = _compute_per_type_projection(
        common_adata=common_adata,
        common_qdata=common_qdata,
        type_names=type_names,
        preliminary_fitted_genes_mask=preliminary_fitted_genes_mask,
        misfit_min_metacells_fraction=misfit_min_metacells_fraction,
        project_log_data=project_log_data,
        project_fold_regularization=project_fold_regularization,
        project_candidates_count=project_candidates_count,
        project_min_significant_gene_umis=project_min_significant_gene_umis,
        project_min_usage_weight=project_min_usage_weight,
        project_max_consistency_fold_factor=project_max_consistency_fold_factor,
        project_max_projection_fold_factor=project_max_projection_fold_factor,
        project_max_projection_noisy_fold_factor=project_max_projection_noisy_fold_factor,
        project_max_misfit_genes=project_max_misfit_genes,
        min_fitted_query_marker_genes=min_fitted_query_marker_genes,
        min_essential_genes_of_type=min_essential_genes_of_type,
        atlas_type_property_name=atlas_type_property_name,
        top_level_parallel=top_level_parallel,
        reproducible=reproducible,
    )

    _compute_dissimilar_residuals_projection(
        weights_per_atlas_per_query_metacell=weights,
        common_adata=common_adata,
        common_qdata=common_qdata,
        project_log_data=project_log_data,
        project_fold_regularization=project_fold_regularization,
        project_candidates_count=project_candidates_count,
        project_min_candidates_fraction=project_min_candidates_fraction,
        project_min_significant_gene_umis=project_min_significant_gene_umis,
        project_min_usage_weight=project_min_usage_weight,
        project_max_consistency_fold_factor=project_max_consistency_fold_factor,
        project_max_projection_fold_factor=project_max_projection_fold_factor,
        project_max_projection_noisy_fold_factor=project_max_projection_noisy_fold_factor,
        project_max_misfit_genes=project_max_misfit_genes,
        min_fitted_query_marker_genes=min_fitted_query_marker_genes,
        min_essential_genes_of_type=min_essential_genes_of_type,
        atlas_type_property_name=atlas_type_property_name,
        top_level_parallel=top_level_parallel,
        reproducible=reproducible,
    )

    tl.compute_projected_fractions(
        adata=common_adata,
        qdata=common_qdata,
        log_data=project_log_data,
        fold_regularization=project_fold_regularization,
        weights=weights,
    )

    _common_data_to_full(
        qdata=qdata,
        common_qdata=common_qdata,
        project_corrections=project_corrections,
        type_names=type_names,
        use_essential_genes=use_essential_genes,
    )

    ut.set_m_data(qdata, "projection_algorithm", f"metacells.{__version__}")
    return sp.csr_matrix(weights)


def _common_gene_indices(adata: AnnData, qdata: AnnData) -> Tuple[ut.NumpyVector, ut.NumpyVector]:
    if list(qdata.var_names) == list(adata.var_names):
        atlas_common_gene_indices = query_common_gene_indices = np.array(range(qdata.n_vars), dtype="int32")
    else:
        atlas_genes_list = list(adata.var_names)
        query_genes_list = list(qdata.var_names)
        common_genes_list = list(sorted(set(atlas_genes_list) & set(query_genes_list)))
        assert len(common_genes_list) > 0
        atlas_common_gene_indices = np.array([atlas_genes_list.index(gene) for gene in common_genes_list])
        query_common_gene_indices = np.array([query_genes_list.index(gene) for gene in common_genes_list])

    return atlas_common_gene_indices, query_common_gene_indices


def _has_any_essential_genes(adata: AnnData) -> bool:
    for property_name in adata.var.keys():
        if property_name.startswith("essential_gene_of_"):
            return True
    return False


def _initialize_atlas_data_in_query(
    adata: AnnData,
    qdata: AnnData,
    *,
    atlas_type_property_name: str,
    atlas_common_gene_indices: ut.NumpyVector,
    query_common_gene_indices: ut.NumpyVector,
    consider_atlas_noisy_genes: bool,
    consider_query_noisy_genes: bool,
    use_essential_genes: bool,
) -> List[str]:
    atlas_genes_mask = np.zeros(qdata.n_vars, dtype="bool")
    atlas_genes_mask[query_common_gene_indices] = True
    ut.set_v_data(qdata, "atlas_gene", atlas_genes_mask)

    type_names = list(np.unique(ut.get_o_numpy(adata, atlas_type_property_name)))

    genes_mask_names = ["lateral_gene", "noisy_gene", "marker_gene"]
    if consider_atlas_noisy_genes:
        genes_mask_names.append("noisy_gene")
    if use_essential_genes:
        genes_mask_names += [f"essential_gene_of_{type_name}" for type_name in type_names]

    for genes_mask_name in genes_mask_names:
        if not ut.has_data(adata, genes_mask_name):
            continue
        atlas_mask = ut.get_v_numpy(adata, genes_mask_name)
        query_mask = np.zeros(qdata.n_vars, dtype="bool")
        query_mask[query_common_gene_indices] = atlas_mask[atlas_common_gene_indices]
        ut.set_v_data(
            qdata,
            genes_mask_name if genes_mask_name.startswith("essential_gene_of_") else "atlas_" + genes_mask_name,
            query_mask,
        )

    noisy_masks: List[str] = []
    if consider_atlas_noisy_genes and ut.has_data(qdata, "atlas_noisy_gene"):
        noisy_masks.append("|atlas_noisy_gene")

    if consider_query_noisy_genes and ut.has_data(qdata, "noisy_gene"):
        noisy_masks.append("|noisy_gene")

    if len(noisy_masks) > 0:
        tl.combine_masks(qdata, noisy_masks, to="projected_noisy_gene")

    return type_names


def _initialize_common_adata(adata: AnnData, what: str, atlas_common_gene_indices: ut.NumpyVector) -> AnnData:
    common_adata = ut.slice(adata, name=".common", vars=atlas_common_gene_indices, top_level=False)

    atlas_total_common_umis = ut.get_o_numpy(common_adata, "total_umis", sum=True)
    ut.set_o_data(common_adata, "total_umis", atlas_total_common_umis)

    _normalize_corrected_fractions(common_adata, what)

    return common_adata


def _initialize_common_qdata(
    qdata: AnnData,
    what: str,
    query_common_gene_indices: ut.NumpyVector,
) -> AnnData:
    common_qdata = ut.slice(
        qdata,
        name=".common",
        vars=query_common_gene_indices,
        track_var="full_gene_index_of_qdata",
        top_level=False,
    )

    query_total_common_umis = ut.get_o_numpy(common_qdata, "total_umis", sum=True)
    ut.set_o_data(common_qdata, "total_umis", query_total_common_umis)
    ut.set_o_data(qdata, "total_atlas_umis", query_total_common_umis)

    _normalize_corrected_fractions(common_qdata, what)

    return common_qdata


def _min_essential_genes_of_type(
    *,
    adata: AnnData,
    common_adata: AnnData,
    type_names: List[str],
    project_min_essential_genes_fraction: Optional[float],
) -> Dict[Tuple[str, str], Optional[int]]:
    min_essential_genes_of_type: Dict[Tuple[str, str], Optional[int]] = {
        (type_name, other_type_name): None for type_name in type_names for other_type_name in type_names
    }

    has_any_essential_genes = False
    if project_min_essential_genes_fraction is not None:
        for type_name in type_names:
            if type_name != "Outliers" and ut.has_data(adata, f"essential_gene_of_{type_name}"):
                has_any_essential_genes = True
                break

    if not has_any_essential_genes:
        return min_essential_genes_of_type
    assert project_min_essential_genes_fraction is not None

    for type_name in type_names:
        if type_name == "Outliers":
            min_essential_genes_count: Optional[int] = None
        else:
            atlas_essential_genes_mask = ut.get_v_numpy(adata, f"essential_gene_of_{type_name}")
            atlas_essential_genes_count = np.sum(atlas_essential_genes_mask)

            common_essential_genes_mask = ut.get_v_numpy(common_adata, f"essential_gene_of_{type_name}")
            common_essential_genes_count = np.sum(common_essential_genes_mask)

            min_essential_genes_count = ceil(project_min_essential_genes_fraction * atlas_essential_genes_count)
            assert min_essential_genes_count is not None
            assert min_essential_genes_count >= 0

            if min_essential_genes_count > common_essential_genes_count:
                atlas_essential_genes_names = sorted(adata.var_names[atlas_essential_genes_mask])
                common_essential_genes_names = sorted(common_adata.var_names[common_essential_genes_mask])
                missing_essential_genes_names = sorted(
                    set(atlas_essential_genes_names) - set(common_essential_genes_names)
                )
                ut.logger().warning(  # pylint: disable=logging-fstring-interpolation
                    f"the {common_essential_genes_count} "
                    f"common essential gene(s) {', '.join(common_essential_genes_names)} "
                    f"for the type {type_name} "
                    f"are not enough for the required {min_essential_genes_count}; "
                    "reducing the minimal requirement "
                    f" (the non-common essential gene(s) are: {', '.join(missing_essential_genes_names)})"
                )
                min_essential_genes_count = common_essential_genes_count

        min_essential_genes_of_type[(type_name, type_name)] = min_essential_genes_count
        ut.log_calc(f"min_essential_genes_of_type[{type_name}]", min_essential_genes_count)

        if min_essential_genes_count is None:
            continue

        for other_type_name in type_names:
            if other_type_name <= type_name or other_type_name == "Outliers":
                continue

            other_atlas_essential_genes_mask = ut.get_v_numpy(adata, f"essential_gene_of_{type_name}")
            pair_atlas_essential_genes_mask = atlas_essential_genes_mask | other_atlas_essential_genes_mask
            pair_atlas_essential_genes_count = np.sum(pair_atlas_essential_genes_mask)

            other_common_essential_genes_mask = ut.get_v_numpy(common_adata, f"essential_gene_of_{type_name}")
            pair_common_essential_genes_mask = common_essential_genes_mask | other_common_essential_genes_mask
            pair_common_essential_genes_count = np.sum(pair_common_essential_genes_mask)

            missing_essential_genes_count = pair_atlas_essential_genes_count - pair_common_essential_genes_count
            assert missing_essential_genes_count >= 0

            min_essential_genes_count = ceil(project_min_essential_genes_fraction * pair_atlas_essential_genes_count)
            assert min_essential_genes_count is not None
            ut.log_calc(f"min_essential_genes_of_type[{type_name}, {other_type_name}]", min_essential_genes_count)
            min_essential_genes_of_type[(type_name, other_type_name)] = min_essential_genes_count
            min_essential_genes_of_type[(other_type_name, type_name)] = min_essential_genes_count

    return min_essential_genes_of_type


def _preliminary_fitted_genes_mask(
    *,
    common_adata: AnnData,
    common_qdata: AnnData,
    only_atlas_marker_genes: bool,
    only_query_marker_genes: bool,
    ignore_atlas_lateral_genes: bool,
    ignore_query_lateral_genes: bool,
    min_significant_gene_umis: int,
) -> ut.NumpyVector:
    atlas_total_umis_per_gene = ut.get_v_numpy(common_adata, "total_umis", sum=True)
    query_total_umis_per_gene = ut.get_v_numpy(common_qdata, "total_umis", sum=True)
    total_umis_per_gene = atlas_total_umis_per_gene + query_total_umis_per_gene
    preliminary_fitted_genes_mask = total_umis_per_gene >= min_significant_gene_umis
    ut.log_calc("total_umis_mask", preliminary_fitted_genes_mask)

    if only_atlas_marker_genes and ut.has_data(common_adata, "marker_gene"):
        preliminary_fitted_genes_mask &= ut.get_v_numpy(common_adata, "marker_gene")

    if only_query_marker_genes and ut.has_data(common_qdata, "marker_gene"):
        preliminary_fitted_genes_mask &= ut.get_v_numpy(common_qdata, "marker_gene")

    if ignore_atlas_lateral_genes and ut.has_data(common_adata, "lateral_gene"):
        preliminary_fitted_genes_mask &= ~ut.get_v_numpy(common_adata, "lateral_gene")

    if ignore_query_lateral_genes and ut.has_data(common_qdata, "lateral_gene"):
        preliminary_fitted_genes_mask &= ~ut.get_v_numpy(common_qdata, "lateral_gene")

    if ut.has_data(common_qdata, "ignored_gene"):
        preliminary_fitted_genes_mask &= ~ut.get_v_numpy(common_qdata, "ignored_gene")

    return preliminary_fitted_genes_mask


def _common_data_to_full(
    *,
    qdata: AnnData,
    common_qdata: AnnData,
    project_corrections: bool,
    use_essential_genes: bool,
    type_names: List[str],
) -> None:
    if use_essential_genes:
        primary_type_per_metacell = ut.get_o_numpy(common_qdata, "projected_type")
        secondary_type_per_metacell = ut.get_o_numpy(common_qdata, "projected_secondary_type")
        essential_per_gene_per_metacell = np.zeros(qdata.shape, dtype="bool")
        essential_gene_per_type = {
            type_name: ut.get_v_numpy(qdata, f"essential_gene_of_{type_name}")
            for type_name in type_names
            if type_name != "Outliers"
        }
        for metacell_index in range(qdata.n_obs):
            primary_type_of_metacell = primary_type_per_metacell[metacell_index]
            if primary_type_of_metacell != "Outliers":
                essential_per_gene_per_metacell[metacell_index, :] = essential_gene_per_type[primary_type_of_metacell]

            secondary_type_of_metacell = secondary_type_per_metacell[metacell_index]
            if secondary_type_of_metacell not in ("", "Outliers", primary_type_of_metacell):
                essential_per_gene_per_metacell[metacell_index, :] |= essential_gene_per_type[
                    secondary_type_of_metacell
                ]

        ut.set_vo_data(qdata, "essential", sp.csr_matrix(essential_per_gene_per_metacell))

    full_gene_index_of_common_qdata = ut.get_v_numpy(common_qdata, "full_gene_index_of_qdata")

    for property_name in (
        "corrected_fraction",
        "projected_fraction",
        "fitted",
        "misfit",
        "projected_fold",
    ):
        data_per_common_gene_per_metacell = ut.get_vo_proper(common_qdata, property_name)
        data_per_gene_per_metacell = np.zeros(qdata.shape, dtype=ut.shaped_dtype(data_per_common_gene_per_metacell))
        data_per_gene_per_metacell[:, full_gene_index_of_common_qdata] = ut.to_numpy_matrix(
            data_per_common_gene_per_metacell
        )
        ut.set_vo_data(qdata, property_name, sp.csr_matrix(data_per_gene_per_metacell))

    property_names = [f"fitted_gene_of_{type_name}" for type_name in type_names]
    if project_corrections:
        property_names.append("correction_factor")

    for property_name in property_names:
        data_per_common_gene = ut.get_v_numpy(common_qdata, property_name)
        data_per_full_gene = np.zeros(qdata.n_vars, dtype=data_per_common_gene.dtype)
        data_per_full_gene[full_gene_index_of_common_qdata] = data_per_common_gene
        ut.set_v_data(qdata, property_name, data_per_full_gene)

    for property_name, formatter in (
        ("projected_type", None),
        ("projected_secondary_type", None),
        ("projected_correlation", ut.sizes_description),
        ("similar", None),
    ):
        data_per_metacell = ut.get_o_numpy(common_qdata, property_name, formatter=formatter)
        ut.set_o_data(qdata, property_name, data_per_metacell, formatter=formatter)


@ut.logged()
@ut.timed_call()
def _compute_preliminary_projection(
    *,
    common_adata: AnnData,
    common_qdata: AnnData,
    preliminary_fitted_genes_mask: ut.NumpyVector,
    project_log_data: bool,
    project_fold_regularization: float,
    project_min_significant_gene_umis: float,
    project_filter_ranges: bool,
    project_ignore_range_quantile: float,
    project_ignore_range_min_overlap_fraction: float,
    project_max_consistency_fold_factor: float,
    project_candidates_count: int,
    project_min_usage_weight: float,
    project_corrections: bool,
    project_min_corrected_gene_correlation: float,
    project_min_corrected_gene_factor: float,
    atlas_type_property_name: str,
    reproducible: bool,
) -> None:
    correction_factor_per_gene = np.full(common_qdata.n_vars, 1.0, dtype="float32")

    repeat = 0
    while True:
        repeat += 1
        ut.log_calc("preliminary repeat", repeat)

        fitted_adata, weights = _compute_correction_factors(
            common_adata=common_adata,
            common_qdata=common_qdata,
            correction_factor_per_gene=correction_factor_per_gene,
            preliminary_fitted_genes_mask=preliminary_fitted_genes_mask,
            project_log_data=project_log_data,
            project_fold_regularization=project_fold_regularization,
            project_min_significant_gene_umis=project_min_significant_gene_umis,
            project_max_consistency_fold_factor=project_max_consistency_fold_factor,
            project_candidates_count=project_candidates_count,
            project_min_usage_weight=project_min_usage_weight,
            project_corrections=project_corrections,
            project_min_corrected_gene_correlation=project_min_corrected_gene_correlation,
            project_min_corrected_gene_factor=project_min_corrected_gene_factor,
            reproducible=reproducible,
        )

        if (
            repeat > 2
            or not project_filter_ranges
            or not _filter_range_genes(
                common_qdata=common_qdata,
                project_fold_regularization=project_fold_regularization,
                project_ignore_range_quantile=project_ignore_range_quantile,
                project_ignore_range_min_overlap_fraction=project_ignore_range_min_overlap_fraction,
                preliminary_fitted_genes_mask=preliminary_fitted_genes_mask,
            )
        ):
            ut.log_calc("preliminary last repeat", repeat)
            break

    if project_corrections:
        ut.set_v_data(common_qdata, "correction_factor", correction_factor_per_gene)

    tl.convey_atlas_to_query(
        adata=fitted_adata,
        qdata=common_qdata,
        weights=weights,
        property_name=atlas_type_property_name,
        to_property_name="projected_type",
    )


@ut.logged()
@ut.timed_call()
def _compute_correction_factors(
    *,
    common_adata: AnnData,
    common_qdata: AnnData,
    correction_factor_per_gene: ut.NumpyVector,
    preliminary_fitted_genes_mask: ut.NumpyVector,
    project_log_data: bool,
    project_fold_regularization: float,
    project_min_significant_gene_umis: float,
    project_max_consistency_fold_factor: float,
    project_candidates_count: int,
    project_min_usage_weight: float,
    project_corrections: bool,
    project_min_corrected_gene_correlation: float,
    project_min_corrected_gene_factor: float,
    reproducible: bool,
) -> Tuple[AnnData, ut.ProperMatrix]:
    repeat = 0
    while True:
        repeat += 1
        ut.log_calc("corrections repeat", repeat)

        fitted_adata = ut.slice(common_adata, name=".fitted", vars=preliminary_fitted_genes_mask, top_level=False)

        fitted_qdata = ut.slice(
            common_qdata,
            name=".fitted",
            vars=preliminary_fitted_genes_mask,
            track_var="common_gene_index_of_qdata",
            top_level=False,
        )

        weights = tl.compute_projection_weights(
            adata=fitted_adata,
            qdata=fitted_qdata,
            log_data=project_log_data,
            fold_regularization=project_fold_regularization,
            min_significant_gene_umis=project_min_significant_gene_umis,
            max_consistency_fold_factor=project_max_consistency_fold_factor,
            candidates_count=project_candidates_count,
            min_usage_weight=project_min_usage_weight,
            reproducible=reproducible,
        )

        tl.compute_projected_fractions(
            adata=common_adata,
            qdata=common_qdata,
            log_data=project_log_data,
            fold_regularization=project_fold_regularization,
            weights=weights,
        )

        if (
            repeat > 2
            or not project_corrections
            or not _correct_correlated_genes(
                common_adata=common_adata,
                common_qdata=common_qdata,
                preliminary_fitted_genes_mask=preliminary_fitted_genes_mask,
                project_min_corrected_gene_correlation=project_min_corrected_gene_correlation,
                project_min_corrected_gene_factor=project_min_corrected_gene_factor,
                correction_factor_per_gene=correction_factor_per_gene,
                reproducible=reproducible,
            )
        ):
            ut.log_calc("corrections last repeat", repeat)
            return fitted_adata, weights


def _correct_correlated_genes(
    *,
    common_adata: AnnData,
    common_qdata: AnnData,
    preliminary_fitted_genes_mask: ut.NumpyVector,
    project_min_corrected_gene_correlation: float,
    project_min_corrected_gene_factor: float,
    correction_factor_per_gene: ut.NumpyVector,
    reproducible: bool,
) -> bool:
    corrected_fractions_per_gene_per_metacell = ut.to_numpy_matrix(
        ut.get_vo_proper(common_qdata, "corrected_fraction", layout="column_major")
    )

    projected_fractions_per_gene_per_metacell = ut.to_numpy_matrix(
        ut.get_vo_proper(common_qdata, "projected_fraction", layout="column_major")
    )

    preliminary_fitted_genes_indices = np.where(preliminary_fitted_genes_mask)[0]
    corrected_fractions_per_fitted_gene_per_metacell = corrected_fractions_per_gene_per_metacell[
        :, preliminary_fitted_genes_indices
    ]
    projected_fractions_per_fitted_gene_per_metacell = projected_fractions_per_gene_per_metacell[
        :, preliminary_fitted_genes_indices
    ]

    correlation_per_fitted_gene = np.full(len(preliminary_fitted_genes_indices), -2, dtype="float32")
    correlation_per_fitted_gene = ut.pairs_corrcoef_rows(
        corrected_fractions_per_fitted_gene_per_metacell.transpose(),
        projected_fractions_per_fitted_gene_per_metacell.transpose(),
        reproducible=reproducible,
    )

    correlated_fitted_genes_mask = correlation_per_fitted_gene >= project_min_corrected_gene_correlation
    ut.log_calc("correlated_fitted_genes_mask", correlated_fitted_genes_mask)
    if not np.any(correlated_fitted_genes_mask):
        return False

    total_corrected_fractions_per_fitted_gene = ut.sum_per(
        corrected_fractions_per_fitted_gene_per_metacell, per="column"
    )
    total_projected_fractions_per_fitted_gene = ut.sum_per(
        projected_fractions_per_fitted_gene_per_metacell, per="column"
    )

    zero_fitted_genes_mask = (total_projected_fractions_per_fitted_gene == 0) | (
        total_corrected_fractions_per_fitted_gene == 0
    )
    ut.log_calc("zero_fitted_genes_mask", zero_fitted_genes_mask)
    total_corrected_fractions_per_fitted_gene[zero_fitted_genes_mask] = 1.0
    total_projected_fractions_per_fitted_gene[zero_fitted_genes_mask] = 1.0

    current_correction_factor_per_fitted_gene = (
        total_projected_fractions_per_fitted_gene / total_corrected_fractions_per_fitted_gene
    )

    high_factor = 1 + project_min_corrected_gene_factor
    low_factor = 1.0 / high_factor
    factor_fitted_genes_mask = (current_correction_factor_per_fitted_gene <= low_factor) | (
        current_correction_factor_per_fitted_gene >= high_factor
    )
    factor_genes_mask = np.zeros(common_adata.n_vars, dtype="bool")
    factor_genes_mask[preliminary_fitted_genes_indices] = factor_fitted_genes_mask
    ut.log_calc("factor_fitted_genes_mask", factor_fitted_genes_mask)
    if not np.any(factor_fitted_genes_mask):
        return False

    corrected_fitted_genes_mask = correlated_fitted_genes_mask & factor_fitted_genes_mask
    ut.log_calc("corrected_fitted_genes_mask", corrected_fitted_genes_mask)
    if not np.any(corrected_fitted_genes_mask):
        return False

    corrected_genes_mask = np.zeros(common_adata.n_vars, dtype="bool")
    corrected_genes_mask[preliminary_fitted_genes_indices] = corrected_fitted_genes_mask
    ut.log_calc("corrected_genes_mask", corrected_genes_mask)

    correction_factor_per_corrected_gene = current_correction_factor_per_fitted_gene[corrected_fitted_genes_mask]
    correction_factor_per_gene[corrected_genes_mask] *= correction_factor_per_corrected_gene

    corrected_fractions_per_gene_per_metacell[:, corrected_genes_mask] *= correction_factor_per_corrected_gene[
        np.newaxis, :
    ]
    corrected_fractions_per_gene_per_metacell = ut.fraction_by(  # type: ignore
        ut.to_layout(corrected_fractions_per_gene_per_metacell, layout="row_major"), by="row"
    )
    ut.set_vo_data(common_qdata, "corrected_fraction", sp.csr_matrix(corrected_fractions_per_gene_per_metacell))

    return True


@ut.logged()
@ut.timed_call()
def _filter_range_genes(
    *,
    common_qdata: AnnData,
    project_fold_regularization: float,
    project_ignore_range_quantile: float,
    project_ignore_range_min_overlap_fraction: float,
    preliminary_fitted_genes_mask: ut.NumpyVector,
) -> bool:
    corrected_fractions_per_gene_per_metacell = ut.to_numpy_matrix(
        ut.get_vo_proper(common_qdata, "corrected_fraction", layout="column_major")
    )
    projected_fractions_per_gene_per_metacell = ut.to_numpy_matrix(
        ut.get_vo_proper(common_qdata, "projected_fraction", layout="column_major")
    )

    preliminary_fitted_genes_indices = np.where(preliminary_fitted_genes_mask)[0]

    corrected_fractions_per_fitted_gene_per_metacell = corrected_fractions_per_gene_per_metacell[
        :, preliminary_fitted_genes_indices
    ]
    projected_fractions_per_fitted_gene_per_metacell = projected_fractions_per_gene_per_metacell[
        :, preliminary_fitted_genes_indices
    ]

    corrected_log_fractions_per_fitted_gene_per_metacell = (
        corrected_fractions_per_fitted_gene_per_metacell + project_fold_regularization
    )
    projected_log_fractions_per_fitted_gene_per_metacell = (
        projected_fractions_per_fitted_gene_per_metacell + project_fold_regularization
    )

    np.log2(
        corrected_log_fractions_per_fitted_gene_per_metacell, out=corrected_log_fractions_per_fitted_gene_per_metacell
    )
    np.log2(
        projected_log_fractions_per_fitted_gene_per_metacell, out=projected_log_fractions_per_fitted_gene_per_metacell
    )

    low_corrected_log_fractions_per_fitted_gene = ut.quantile_per(
        corrected_log_fractions_per_fitted_gene_per_metacell, project_ignore_range_quantile, per="column"
    )
    low_projected_log_fractions_per_fitted_gene = ut.quantile_per(
        projected_log_fractions_per_fitted_gene_per_metacell, project_ignore_range_quantile, per="column"
    )

    high_corrected_log_fractions_per_fitted_gene = ut.quantile_per(
        corrected_log_fractions_per_fitted_gene_per_metacell, 1.0 - project_ignore_range_quantile, per="column"
    )
    high_projected_log_fractions_per_fitted_gene = ut.quantile_per(
        projected_log_fractions_per_fitted_gene_per_metacell, 1.0 - project_ignore_range_quantile, per="column"
    )

    low_common_log_fractions_per_fitted_gene = np.maximum(
        low_corrected_log_fractions_per_fitted_gene, low_projected_log_fractions_per_fitted_gene
    )
    high_common_log_fractions_per_fitted_gene = np.minimum(
        high_corrected_log_fractions_per_fitted_gene, high_projected_log_fractions_per_fitted_gene
    )

    corrected_range_log_fractions_per_fitted_gene = (
        high_corrected_log_fractions_per_fitted_gene - low_corrected_log_fractions_per_fitted_gene
    )
    common_range_log_fractions_per_fitted_gene = (
        high_common_log_fractions_per_fitted_gene - low_common_log_fractions_per_fitted_gene
    )

    corrected_range_log_fractions_per_fitted_gene[corrected_range_log_fractions_per_fitted_gene == 0] = 1
    overlap_per_fitted_gene = common_range_log_fractions_per_fitted_gene / corrected_range_log_fractions_per_fitted_gene

    ignore_fitted_genes_mask = overlap_per_fitted_gene < project_ignore_range_min_overlap_fraction
    ut.log_calc("ignore_fitted_genes_mask", ignore_fitted_genes_mask)
    if not np.any(ignore_fitted_genes_mask):
        return False

    ignore_gene_indices = preliminary_fitted_genes_indices[ignore_fitted_genes_mask]
    preliminary_fitted_genes_mask[ignore_gene_indices] = False
    ut.log_calc("preliminary_fitted_genes_mask", preliminary_fitted_genes_mask)

    return True


@ut.logged()
@ut.timed_call()
def _compute_per_type_projection(
    *,
    common_adata: AnnData,
    common_qdata: AnnData,
    type_names: List[str],
    preliminary_fitted_genes_mask: ut.NumpyVector,
    misfit_min_metacells_fraction: float,
    project_log_data: bool,
    project_fold_regularization: float,
    project_candidates_count: int,
    project_min_significant_gene_umis: float,
    project_min_usage_weight: float,
    project_max_consistency_fold_factor: float,
    project_max_projection_fold_factor: float,
    project_max_projection_noisy_fold_factor: float,
    project_max_misfit_genes: int,
    min_fitted_query_marker_genes: float,
    min_essential_genes_of_type: Dict[Tuple[str, str], Optional[int]],
    atlas_type_property_name: str,
    top_level_parallel: bool,
    reproducible: bool,
) -> ut.NumpyMatrix:
    old_types_per_metacell: List[Set[str]] = []
    for _metacell_index in range(common_qdata.n_obs):
        old_types_per_metacell.append(set())

    fitted_genes_mask_per_type = _initial_fitted_genes_mask_per_type(
        common_qdata=common_qdata, type_names=type_names, preliminary_fitted_genes_mask=preliminary_fitted_genes_mask
    )

    misfit_per_gene_per_metacell = np.empty(common_qdata.shape, dtype="bool")
    projected_correlation_per_metacell = np.empty(common_qdata.n_obs, dtype="float32")
    projected_fold_per_gene_per_metacell = np.empty(common_qdata.shape, dtype="float32")
    weights_per_atlas_per_query_metacell = np.empty((common_qdata.n_obs, common_adata.n_obs), dtype="float32")
    similar_per_metacell = np.empty(common_qdata.n_obs, dtype="bool")

    repeat = 0
    while True:
        repeat += 1
        ut.log_calc("types repeat", repeat)

        type_of_query_metacells = ut.get_o_numpy(common_qdata, "projected_type")
        query_type_names = list(np.unique(type_of_query_metacells))

        misfit_per_gene_per_metacell[:, :] = False
        projected_fold_per_gene_per_metacell[:, :] = 0.0

        @ut.timed_call("single_type_projection")
        def _single_type_projection(
            type_index: int,
        ) -> Dict[str, Any]:
            return _compute_single_type_projection(
                type_name=query_type_names[type_index],
                common_adata=common_adata,
                common_qdata=common_qdata,
                fitted_genes_mask_per_type=fitted_genes_mask_per_type,
                misfit_min_metacells_fraction=misfit_min_metacells_fraction,
                project_log_data=project_log_data,
                project_fold_regularization=project_fold_regularization,
                project_candidates_count=project_candidates_count,
                project_min_significant_gene_umis=project_min_significant_gene_umis,
                project_min_usage_weight=project_min_usage_weight,
                project_max_consistency_fold_factor=project_max_consistency_fold_factor,
                project_max_projection_fold_factor=project_max_projection_fold_factor,
                project_max_projection_noisy_fold_factor=project_max_projection_noisy_fold_factor,
                project_max_misfit_genes=project_max_misfit_genes,
                min_fitted_query_marker_genes=min_fitted_query_marker_genes,
                min_essential_genes_of_type=min_essential_genes_of_type,
                top_level_parallel=top_level_parallel,
                reproducible=reproducible,
            )

        @ut.logged()
        def _collect_single_type_result(
            type_index: int,
            *,
            query_metacell_indices_of_type: ut.NumpyVector,
            fitted_genes_indices_of_type: ut.NumpyVector,
            similar_per_metacell_of_type: ut.NumpyVector,
            misfit_per_gene_per_metacell_of_type: ut.ProperMatrix,
            projected_correlation_per_metacell_of_type: ut.NumpyVector,
            projected_fold_per_gene_per_metacell_of_type: ut.ProperMatrix,
            weights_per_atlas_per_query_metacell_of_type: ut.ProperMatrix,
        ) -> None:
            fitted_genes_mask_per_type[query_type_names[type_index]][:] = False
            fitted_genes_mask_per_type[query_type_names[type_index]][fitted_genes_indices_of_type] = True
            similar_per_metacell[query_metacell_indices_of_type] = similar_per_metacell_of_type
            projected_correlation_per_metacell[
                query_metacell_indices_of_type
            ] = projected_correlation_per_metacell_of_type
            misfit_per_gene_per_metacell[query_metacell_indices_of_type, :] = ut.to_numpy_matrix(
                misfit_per_gene_per_metacell_of_type
            )
            projected_fold_per_gene_per_metacell[query_metacell_indices_of_type, :] = ut.to_numpy_matrix(
                projected_fold_per_gene_per_metacell_of_type
            )
            weights_per_atlas_per_query_metacell[query_metacell_indices_of_type, :] = ut.to_numpy_matrix(
                weights_per_atlas_per_query_metacell_of_type
            )

        if top_level_parallel:
            for type_index, result in enumerate(ut.parallel_map(_single_type_projection, len(query_type_names))):
                _collect_single_type_result(type_index, **result)
        else:
            for type_index in range(len(query_type_names)):
                result = _single_type_projection(type_index)
                _collect_single_type_result(type_index, **result)

        if repeat > 2 or not _changed_projected_types(
            common_adata=common_adata,
            common_qdata=common_qdata,
            old_type_of_query_metacells=type_of_query_metacells,
            weights_per_atlas_per_query_metacell=weights_per_atlas_per_query_metacell,
            atlas_type_property_name=atlas_type_property_name,
            old_types_per_metacell=old_types_per_metacell,
        ):
            ut.log_calc("types last repeat", repeat)
            break

    for type_name, fitted_genes_mask_of_type in fitted_genes_mask_per_type.items():
        ut.set_v_data(common_qdata, f"fitted_gene_of_{type_name}", fitted_genes_mask_of_type)

    projected_type_per_metacell = ut.get_o_numpy(common_qdata, "projected_type")
    fitted_mask_per_gene_per_metacell = np.vstack(
        [fitted_genes_mask_per_type[type_name] for type_name in projected_type_per_metacell]
    )
    ut.set_vo_data(common_qdata, "fitted", fitted_mask_per_gene_per_metacell)

    ut.set_o_data(common_qdata, "similar", similar_per_metacell)
    ut.set_o_data(common_qdata, "projected_correlation", projected_correlation_per_metacell)

    ut.set_vo_data(common_qdata, "misfit", misfit_per_gene_per_metacell)
    ut.set_vo_data(common_qdata, "projected_fold", projected_fold_per_gene_per_metacell)

    tl.compute_projected_fractions(
        adata=common_adata,
        qdata=common_qdata,
        log_data=project_log_data,
        fold_regularization=project_fold_regularization,
        weights=weights_per_atlas_per_query_metacell,
    )

    return weights_per_atlas_per_query_metacell


def _initial_fitted_genes_mask_per_type(
    common_qdata: AnnData,
    type_names: List[str],
    preliminary_fitted_genes_mask: ut.NumpyVector,
) -> Dict[str, ut.NumpyVector]:
    fitted_genes_mask_per_type: Dict[str, ut.NumpyVector] = {}
    for type_name in type_names:
        fitted_genes_mask_of_type = preliminary_fitted_genes_mask.copy()
        property_name = f"ignored_gene_of_{type_name}"
        if ut.has_data(common_qdata, property_name):
            ignored_gene_mask_of_type = ut.get_v_numpy(common_qdata, property_name)
            fitted_genes_mask_of_type &= ~ignored_gene_mask_of_type
        fitted_genes_mask_per_type[type_name] = fitted_genes_mask_of_type
    return fitted_genes_mask_per_type


@ut.logged()
@ut.timed_call()
def _compute_single_type_projection(
    *,
    type_name: str,
    common_adata: AnnData,
    common_qdata: AnnData,
    fitted_genes_mask_per_type: Dict[str, ut.NumpyVector],
    misfit_min_metacells_fraction: float,
    project_log_data: bool,
    project_fold_regularization: float,
    project_candidates_count: int,
    project_min_significant_gene_umis: float,
    project_min_usage_weight: float,
    project_max_consistency_fold_factor: float,
    project_max_projection_fold_factor: float,
    project_max_projection_noisy_fold_factor: float,
    project_max_misfit_genes: int,
    min_fitted_query_marker_genes: float,
    min_essential_genes_of_type: Dict[Tuple[str, str], Optional[int]],
    top_level_parallel: bool,
    reproducible: bool,
) -> Dict[str, Any]:
    projected_type_per_metacell = ut.get_o_numpy(common_qdata, "projected_type")
    query_metacell_mask_of_type = projected_type_per_metacell == type_name
    ut.log_calc("query_metacell_mask_of_type", query_metacell_mask_of_type)
    assert np.any(query_metacell_mask_of_type)
    type_common_qdata = ut.slice(common_qdata, name=f".{type_name}", obs=query_metacell_mask_of_type, top_level=False)
    corrected_fractions = ut.get_vo_proper(type_common_qdata, "corrected_fraction")

    fitted_genes_mask_of_type = fitted_genes_mask_per_type[type_name]
    assert np.any(fitted_genes_mask_of_type)

    repeat = 0
    while True:
        repeat += 1
        ut.log_calc(f"{type_name} misfit repeat", repeat)

        type_fitted_adata = ut.slice(
            common_adata, name=f".{type_name}.fitted", vars=fitted_genes_mask_of_type, top_level=False
        )

        type_fitted_qdata = ut.slice(
            type_common_qdata,
            name=".fitted",
            vars=fitted_genes_mask_of_type,
            track_var="common_gene_index_of_qdata",
            top_level=False,
        )

        type_weights = tl.compute_projection_weights(
            adata=type_fitted_adata,
            qdata=type_fitted_qdata,
            log_data=project_log_data,
            fold_regularization=project_fold_regularization,
            min_significant_gene_umis=project_min_significant_gene_umis,
            max_consistency_fold_factor=project_max_consistency_fold_factor,
            candidates_count=project_candidates_count,
            min_usage_weight=project_min_usage_weight,
            reproducible=reproducible,
        )

        tl.compute_projected_fractions(
            adata=common_adata,
            qdata=type_common_qdata,
            log_data=project_log_data,
            fold_regularization=project_fold_regularization,
            weights=type_weights,
        )
        projected_fractions = ut.get_vo_proper(type_common_qdata, "projected_fraction")

        tl.compute_projected_folds(
            type_common_qdata,
            fold_regularization=project_fold_regularization,
            min_significant_gene_umis=project_min_significant_gene_umis,
        )
        projected_fold_per_gene_per_metacell_of_type = ut.get_vo_proper(type_common_qdata, "projected_fold")

        if not _detect_type_misfit_genes(
            type_common_qdata=type_common_qdata,
            projected_fold_per_gene_per_metacell_of_type=projected_fold_per_gene_per_metacell_of_type,
            max_projection_fold_factor=project_max_projection_fold_factor,
            max_projection_noisy_fold_factor=project_max_projection_noisy_fold_factor,
            misfit_min_metacells_fraction=misfit_min_metacells_fraction,
            fitted_genes_mask_of_type=fitted_genes_mask_of_type,
        ):
            ut.log_calc(f"{type_name} misfit last repeat", repeat)
            break

    tl.compute_similar_query_metacells(
        type_common_qdata,
        max_projection_fold_factor=project_max_projection_fold_factor,
        max_projection_noisy_fold_factor=project_max_projection_noisy_fold_factor,
        max_misfit_genes=project_max_misfit_genes,
        min_fitted_query_marker_genes=min_fitted_query_marker_genes,
        essential_genes_property=f"essential_gene_of_{type_name}",
        min_essential_genes=min_essential_genes_of_type[(type_name, type_name)],
        fitted_genes_mask=fitted_genes_mask_of_type,
    )
    similar_per_metacell_of_type = ut.get_o_numpy(type_common_qdata, "similar")
    misfit_per_gene_per_metacell_of_type = ut.get_vo_proper(type_common_qdata, "misfit")

    fitted_corrected_fractions = corrected_fractions[:, fitted_genes_mask_of_type]
    fitted_projected_fractions = projected_fractions[:, fitted_genes_mask_of_type]
    projected_correlation_per_metacell_of_type = ut.pairs_corrcoef_rows(
        ut.to_layout(ut.to_numpy_matrix(fitted_corrected_fractions), layout="row_major"),
        ut.to_layout(ut.to_numpy_matrix(fitted_projected_fractions), layout="row_major"),
        reproducible=reproducible,
    )

    if top_level_parallel:
        if not isinstance(misfit_per_gene_per_metacell_of_type, sp.csr_matrix):
            misfit_per_gene_per_metacell_of_type = sp.csr_matrix(misfit_per_gene_per_metacell_of_type)
        if not isinstance(projected_fold_per_gene_per_metacell_of_type, sp.csr_matrix):
            projected_fold_per_gene_per_metacell_of_type = sp.csr_matrix(projected_fold_per_gene_per_metacell_of_type)
        if not isinstance(type_weights, sp.csr_matrix):
            type_weights = sp.csr_matrix(type_weights)

    ut.log_return("query_metacell_mask_of_type", query_metacell_mask_of_type)
    ut.log_return("fitted_genes_mask_of_type", fitted_genes_mask_of_type)
    ut.log_return("similar_per_metacell_of_type", similar_per_metacell_of_type)
    ut.log_return("misfit_per_gene_per_metacell_of_type", misfit_per_gene_per_metacell_of_type)
    ut.log_return("projected_correlation_per_metacell_of_type", projected_correlation_per_metacell_of_type)
    ut.log_return("projected_fold_per_gene_per_metacell_of_type", projected_fold_per_gene_per_metacell_of_type)
    ut.log_return("weights", type_weights)

    return {
        "query_metacell_indices_of_type": np.where(query_metacell_mask_of_type)[0],
        "fitted_genes_indices_of_type": np.where(fitted_genes_mask_of_type)[0],
        "similar_per_metacell_of_type": similar_per_metacell_of_type,
        "misfit_per_gene_per_metacell_of_type": misfit_per_gene_per_metacell_of_type,
        "projected_correlation_per_metacell_of_type": projected_correlation_per_metacell_of_type,
        "projected_fold_per_gene_per_metacell_of_type": projected_fold_per_gene_per_metacell_of_type,
        "weights_per_atlas_per_query_metacell_of_type": type_weights,
    }


def _detect_type_misfit_genes(
    *,
    type_common_qdata: AnnData,
    projected_fold_per_gene_per_metacell_of_type: ut.ProperMatrix,
    max_projection_fold_factor: float,
    max_projection_noisy_fold_factor: float,
    misfit_min_metacells_fraction: float,
    fitted_genes_mask_of_type: ut.NumpyVector,
) -> bool:
    assert max_projection_fold_factor >= 0
    assert max_projection_noisy_fold_factor >= 0
    assert 0 <= misfit_min_metacells_fraction <= 1

    if ut.has_data(type_common_qdata, "projected_noisy_gene"):
        max_projection_fold_factor_per_gene = np.full(
            type_common_qdata.n_vars, max_projection_fold_factor, dtype="float32"
        )
        noisy_per_gene = ut.get_v_numpy(type_common_qdata, "projected_noisy_gene")
        max_projection_fold_factor_per_gene[noisy_per_gene] += max_projection_noisy_fold_factor
        high_projection_fold_per_gene_per_metacell_of_type = (
            projected_fold_per_gene_per_metacell_of_type > max_projection_fold_factor_per_gene[np.newaxis, :]
        )
    else:
        high_projection_fold_per_gene_per_metacell_of_type = (
            ut.to_numpy_matrix(projected_fold_per_gene_per_metacell_of_type) > max_projection_fold_factor
        )
    ut.log_calc(
        "high_projection_fold_per_gene_per_metacell_of_type", high_projection_fold_per_gene_per_metacell_of_type
    )

    high_projection_metacells_per_gene = ut.sum_per(
        ut.to_layout(high_projection_fold_per_gene_per_metacell_of_type, layout="column_major"), per="column"
    )
    ut.log_calc(
        "high_projection_metacells_per_gene", high_projection_metacells_per_gene, formatter=ut.sizes_description
    )

    min_high_projection_metacells = type_common_qdata.n_obs * misfit_min_metacells_fraction
    ut.log_calc("min_high_projection_metacells", min_high_projection_metacells)

    high_projection_genes_mask = high_projection_metacells_per_gene >= min_high_projection_metacells
    type_misfit_genes_mask = fitted_genes_mask_of_type & high_projection_genes_mask
    ut.log_calc("type_misfit_genes_mask", type_misfit_genes_mask)
    if not np.any(type_misfit_genes_mask):
        return False

    fitted_genes_mask_of_type[type_misfit_genes_mask] = False
    ut.log_calc("fitted_genes_mask_of_type", fitted_genes_mask_of_type)

    return True


@ut.logged()
@ut.timed_call()
def _changed_projected_types(
    *,
    common_adata: AnnData,
    common_qdata: AnnData,
    old_type_of_query_metacells: ut.NumpyVector,
    weights_per_atlas_per_query_metacell: ut.NumpyMatrix,
    atlas_type_property_name: str,
    old_types_per_metacell: List[Set[str]],
) -> bool:
    tl.convey_atlas_to_query(
        adata=common_adata,
        qdata=common_qdata,
        weights=weights_per_atlas_per_query_metacell,
        property_name=atlas_type_property_name,
        to_property_name="projected_type",
    )
    new_type_of_query_metacells = ut.get_o_numpy(common_qdata, "projected_type")

    has_changed = False
    for metacell_index, old_types_of_metacell in enumerate(old_types_per_metacell):
        old_type = old_type_of_query_metacells[metacell_index]
        old_types_of_metacell.add(old_type)

        new_type = new_type_of_query_metacells[metacell_index]
        if new_type not in old_types_of_metacell:
            ut.log_calc(f"metacell: {metacell_index} changed from old type: {old_type} to new type", new_type)
            has_changed = True
        elif new_type != old_type:
            ut.log_calc(f"metacell: {metacell_index} changed from old type: {old_type} to older type", new_type)

    ut.log_return("has_changed", has_changed)
    return has_changed


@ut.timed_call()
@ut.logged()
def _compute_dissimilar_residuals_projection(
    *,
    common_adata: AnnData,
    common_qdata: AnnData,
    weights_per_atlas_per_query_metacell: ut.NumpyMatrix,
    project_log_data: bool,
    project_fold_regularization: float,
    project_candidates_count: int,
    project_min_candidates_fraction: float,
    project_min_significant_gene_umis: float,
    project_min_usage_weight: float,
    project_max_consistency_fold_factor: float,
    project_max_projection_fold_factor: float,
    project_max_projection_noisy_fold_factor: float,
    project_max_misfit_genes: int,
    min_fitted_query_marker_genes: float,
    min_essential_genes_of_type: Dict[Tuple[str, str], Optional[int]],
    atlas_type_property_name: str,
    top_level_parallel: bool,
    reproducible: bool,
) -> None:
    secondary_type = [""] * common_qdata.n_obs
    dissimilar_metacells_mask = ~ut.get_o_numpy(common_qdata, "similar")
    if not np.any(dissimilar_metacells_mask):
        ut.set_o_data(common_qdata, "projected_secondary_type", np.array(secondary_type))
        return
    dissimilar_metacell_indices = np.where(dissimilar_metacells_mask)[0]

    @ut.timed_call("single_metacell_residuals")
    def _single_metacell_residuals(
        dissimilar_metacell_position: int,
    ) -> Optional[Dict[str, Any]]:
        return _compute_single_metacell_residuals(
            dissimilar_metacell_index=dissimilar_metacell_indices[dissimilar_metacell_position],
            common_adata=common_adata,
            common_qdata=common_qdata,
            project_log_data=project_log_data,
            project_fold_regularization=project_fold_regularization,
            project_candidates_count=project_candidates_count,
            project_min_candidates_fraction=project_min_candidates_fraction,
            project_min_significant_gene_umis=project_min_significant_gene_umis,
            project_min_usage_weight=project_min_usage_weight,
            project_max_consistency_fold_factor=project_max_consistency_fold_factor,
            project_max_projection_fold_factor=project_max_projection_fold_factor,
            project_max_projection_noisy_fold_factor=project_max_projection_noisy_fold_factor,
            project_max_misfit_genes=project_max_misfit_genes,
            min_fitted_query_marker_genes=min_fitted_query_marker_genes,
            min_essential_genes_of_type=min_essential_genes_of_type,
            atlas_type_property_name=atlas_type_property_name,
            reproducible=reproducible,
        )

    if top_level_parallel:
        results = ut.parallel_map(_single_metacell_residuals, len(dissimilar_metacell_indices))
    else:
        results = [
            _single_metacell_residuals(dissimilar_metacell_position)
            for dissimilar_metacell_position in range(len(dissimilar_metacell_indices))
        ]

    similar_per_metacell = ut.get_o_numpy(common_qdata, "similar").copy()
    primary_type_per_metacell = ut.get_o_numpy(common_qdata, "projected_type").copy()
    secondary_type_per_metacell = [""] * common_qdata.n_obs
    fitted_per_gene_per_metacell = ut.to_numpy_matrix(ut.get_vo_proper(common_qdata, "fitted"), copy=True)
    misfit_per_gene_per_metacell = ut.to_numpy_matrix(ut.get_vo_proper(common_qdata, "misfit"), copy=True)
    projected_correlation_per_metacell = ut.get_o_numpy(common_qdata, "projected_correlation").copy()
    projected_fold_per_gene_per_metacell = ut.to_numpy_matrix(
        ut.get_vo_proper(common_qdata, "projected_fold"), copy=True
    )

    def _collect_metacell_residuals(
        dissimilar_metacell_index: int,
        primary_type: str,
        secondary_type: str,
        fitted_genes_mask: ut.NumpyVector,
        misfit_genes_mask: ut.NumpyVector,
        projected_correlation: float,
        projected_fold_per_gene: ut.NumpyVector,
        weights: ut.ProperMatrix,
    ) -> None:
        similar_per_metacell[dissimilar_metacell_index] = True
        primary_type_per_metacell[dissimilar_metacell_index] = primary_type
        secondary_type_per_metacell[dissimilar_metacell_index] = secondary_type
        projected_correlation_per_metacell[dissimilar_metacell_index] = projected_correlation

        fitted_per_gene_per_metacell[dissimilar_metacell_index, :] = fitted_genes_mask
        misfit_per_gene_per_metacell[dissimilar_metacell_index, :] = misfit_genes_mask
        projected_fold_per_gene_per_metacell[dissimilar_metacell_index, :] = projected_fold_per_gene
        weights_per_atlas_per_query_metacell[dissimilar_metacell_index, :] = weights

    for dissimilar_metacell_index, result in zip(dissimilar_metacell_indices, results):
        if result is not None:
            _collect_metacell_residuals(dissimilar_metacell_index, **result)

    ut.set_o_data(common_qdata, "similar", similar_per_metacell)
    ut.set_o_data(common_qdata, "projected_type", primary_type_per_metacell)
    ut.set_o_data(common_qdata, "projected_secondary_type", np.array(secondary_type_per_metacell))
    ut.set_o_data(common_qdata, "projected_correlation", projected_correlation_per_metacell)

    ut.set_vo_data(common_qdata, "fitted", fitted_per_gene_per_metacell)
    ut.set_vo_data(common_qdata, "misfit", misfit_per_gene_per_metacell)
    ut.set_vo_data(common_qdata, "projected_fold", projected_fold_per_gene_per_metacell)


@ut.timed_call()
@ut.logged()
def _compute_single_metacell_residuals(  # pylint: disable=too-many-statements
    *,
    dissimilar_metacell_index: int,
    common_adata: AnnData,
    common_qdata: AnnData,
    project_log_data: bool,
    project_fold_regularization: float,
    project_candidates_count: int,
    project_min_candidates_fraction: float,
    project_min_significant_gene_umis: float,
    project_min_usage_weight: float,
    project_max_consistency_fold_factor: float,
    project_max_projection_fold_factor: float,
    project_max_projection_noisy_fold_factor: float,
    project_max_misfit_genes: int,
    min_fitted_query_marker_genes: float,
    min_essential_genes_of_type: Dict[Tuple[str, str], Optional[int]],
    atlas_type_property_name: str,
    reproducible: bool,
) -> Optional[Dict[str, Any]]:
    dissimilar_qdata = ut.slice(
        common_qdata, name=f".dissimilar.{dissimilar_metacell_index}", obs=[dissimilar_metacell_index], top_level=False
    )
    corrected_fractions = ut.get_vo_proper(dissimilar_qdata, "corrected_fraction")

    primary_type = ut.get_o_numpy(dissimilar_qdata, "projected_type")[0]
    secondary_type = ""

    repeat = 0
    while True:
        repeat += 1
        ut.log_calc("residuals repeat", repeat)
        ut.log_calc("primary_type", primary_type)
        ut.log_calc("secondary_type", secondary_type)

        fitted_genes_mask = ut.get_v_numpy(dissimilar_qdata, f"fitted_gene_of_{primary_type}")
        if secondary_type != "":
            fitted_genes_mask = fitted_genes_mask | ut.get_v_numpy(dissimilar_qdata, f"fitted_gene_of_{secondary_type}")

        fitted_adata = ut.slice(
            common_adata,
            name=f".dissimilar.{dissimilar_metacell_index}.fitted",
            vars=fitted_genes_mask,
            top_level=False,
        )

        metacell_fitted_qdata = ut.slice(dissimilar_qdata, name=".fitted", vars=fitted_genes_mask, top_level=False)

        second_anchor_indices: List[int] = []
        weights = tl.compute_projection_weights(
            adata=fitted_adata,
            qdata=metacell_fitted_qdata,
            log_data=project_log_data,
            fold_regularization=project_fold_regularization,
            min_significant_gene_umis=project_min_significant_gene_umis,
            max_consistency_fold_factor=project_max_consistency_fold_factor,
            candidates_count=project_candidates_count,
            min_candidates_fraction=project_min_candidates_fraction,
            min_usage_weight=project_min_usage_weight,
            reproducible=reproducible,
            second_anchor_indices=second_anchor_indices,
        )
        tl.compute_projected_fractions(
            adata=common_adata,
            qdata=dissimilar_qdata,
            log_data=project_log_data,
            fold_regularization=project_fold_regularization,
            weights=weights,
        )
        projected_fractions = ut.get_vo_proper(dissimilar_qdata, "projected_fraction")

        first_anchor_weights = weights.copy()
        first_anchor_weights[:, second_anchor_indices] = 0.0

        if len(second_anchor_indices) == 0:
            new_secondary_type = ""
        else:
            second_anchor_weights = weights - first_anchor_weights  # type: ignore

            tl.convey_atlas_to_query(
                adata=common_adata,
                qdata=dissimilar_qdata,
                weights=second_anchor_weights,
                property_name=atlas_type_property_name,
                to_property_name="projected_secondary_type",
            )
            new_secondary_type = ut.get_o_numpy(dissimilar_qdata, "projected_secondary_type")[0]

        if np.sum(first_anchor_weights.data) == 0:
            new_primary_type = new_secondary_type
            new_secondary_type = ""
        else:
            tl.convey_atlas_to_query(
                adata=common_adata,
                qdata=dissimilar_qdata,
                weights=first_anchor_weights,
                property_name=atlas_type_property_name,
                to_property_name="projected_type",
            )
            new_primary_type = ut.get_o_numpy(dissimilar_qdata, "projected_type")[0]

        if repeat > 2 or (new_primary_type == primary_type and new_secondary_type == secondary_type):
            ut.log_calc("residuals last repeat", repeat)
            break

        primary_type = new_primary_type
        secondary_type = new_secondary_type

    projected_correlation = ut.pairs_corrcoef_rows(
        ut.to_numpy_matrix(corrected_fractions), ut.to_numpy_matrix(projected_fractions), reproducible=reproducible
    )[0]

    tl.compute_projected_folds(
        dissimilar_qdata,
        fold_regularization=project_fold_regularization,
        min_significant_gene_umis=project_min_significant_gene_umis,
    )
    projected_fold_per_gene = ut.to_numpy_vector(ut.get_vo_proper(dissimilar_qdata, "projected_fold"))

    essential_genes_properties = [f"essential_gene_of_{primary_type}"]
    if secondary_type == "":
        min_essential_genes = min_essential_genes_of_type[(primary_type, primary_type)]
    else:
        min_essential_genes = min_essential_genes_of_type[(primary_type, secondary_type)]
        essential_genes_properties.append(f"essential_gene_of_{secondary_type}")

    tl.compute_similar_query_metacells(
        dissimilar_qdata,
        max_projection_fold_factor=project_max_projection_fold_factor,
        max_projection_noisy_fold_factor=project_max_projection_noisy_fold_factor,
        max_misfit_genes=project_max_misfit_genes,
        essential_genes_property=essential_genes_properties,
        min_essential_genes=min_essential_genes,
        min_fitted_query_marker_genes=min_fitted_query_marker_genes,
        fitted_genes_mask=fitted_genes_mask,
    )

    similar = ut.get_o_numpy(dissimilar_qdata, "similar")[0]
    ut.log_return("similar", False)
    if not similar:
        return None

    misfit_genes_mask = ut.to_numpy_vector(ut.get_vo_proper(dissimilar_qdata, "misfit")[0, :])

    ut.log_return("primary_type", primary_type)
    ut.log_return("secondary_type", secondary_type)
    ut.log_return("fitted_genes_mask", fitted_genes_mask)
    ut.log_return("misfit_genes_mask", misfit_genes_mask)
    ut.log_return("projected_correlation", projected_correlation)
    ut.log_return("projected_fold_per_gene", projected_fold_per_gene)

    return {
        "primary_type": primary_type,
        "secondary_type": secondary_type,
        "fitted_genes_mask": fitted_genes_mask,
        "misfit_genes_mask": misfit_genes_mask,
        "projected_correlation": projected_correlation,
        "projected_fold_per_gene": projected_fold_per_gene,
        "weights": ut.to_numpy_vector(weights),
    }


def _normalize_corrected_fractions(adata: AnnData, what: str) -> None:
    fractions_data = ut.get_vo_proper(adata, what, layout="row_major")
    corrected_data = ut.fraction_by(fractions_data, by="row")
    ut.set_vo_data(adata, "corrected_fraction", corrected_data)


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def outliers_projection_pipeline(
    what: str = "__x__",
    *,
    adata: AnnData,
    odata: AnnData,
    fold_regularization: float = pr.outliers_fold_regularization,
    project_min_significant_gene_umis: int = pr.project_min_significant_gene_umis,
    reproducible: bool,
) -> None:
    """
    Project outliers on an atlas.

    **Returns**

    Sets the following in ``odata``:

    Per-Observation (Cell) Annotations

        ``atlas_most_similar``
            For each observation (outlier), the index of the "most similar" atlas metacell.

    Per-Variable Per-Observation (Gene-Cell) Annotations

        ``atlas_most_similar_fold``
            The fold factor between the outlier gene expression and their expression in the most similar atlas metacell
            (unless the value is too low to be of interest, in which case it will be zero).

    **Computation Parameters**

    1. Invoke :py:func:`metacells.tools.quality.compute_outliers_matches` using the ``fold_regularization``
       (default: {fold_regularization}) and ``reproducible``.

    2. Invoke :py:func:`metacells.tools.quality.compute_outliers_fold_factors` using the ``fold_regularization``
       (default: {fold_regularization}).
    """
    if list(odata.var_names) != list(adata.var_names):
        atlas_genes_list = list(adata.var_names)
        query_genes_list = list(odata.var_names)
        common_genes_list = list(sorted(set(atlas_genes_list) & set(query_genes_list)))
        atlas_common_gene_indices = np.array([atlas_genes_list.index(gene) for gene in common_genes_list])
        query_common_gene_indices = np.array([query_genes_list.index(gene) for gene in common_genes_list])
        common_adata = ut.slice(adata, name=".common", vars=atlas_common_gene_indices, top_level=False)
        common_odata = ut.slice(
            odata,
            name=".common",
            vars=query_common_gene_indices,
            top_level=False,
        )
    else:
        common_adata = adata
        common_odata = odata

    tl.compute_outliers_matches(
        what,
        adata=common_odata,
        gdata=common_adata,
        most_similar="atlas_most_similar",
        value_regularization=fold_regularization,
        reproducible=reproducible,
    )

    tl.compute_outliers_fold_factors(
        what,
        adata=common_odata,
        gdata=common_adata,
        most_similar="atlas_most_similar",
        min_gene_total=project_min_significant_gene_umis,
    )

    if list(odata.var_names) != list(adata.var_names):
        atlas_most_similar = ut.get_o_numpy(common_odata, "atlas_most_similar")
        ut.set_o_data(odata, "atlas_most_similar", atlas_most_similar)

        common_folds = ut.get_vo_proper(common_odata, "atlas_most_similar_fold")
        atlas_most_similar_fold = np.zeros(odata.shape, dtype="float32")
        atlas_most_similar_fold[:, query_common_gene_indices] = common_folds
        ut.set_vo_data(odata, "atlas_most_similar_fold", sp.csr_matrix(atlas_most_similar_fold))


def write_projection_weights(path: str, adata: AnnData, qdata: AnnData, weights: ut.CompressedMatrix) -> None:
    """
    Write into the ``path`` the ``weights`` computed for the projection of the query ``qdata`` on the atlas ``adata``.

    Since the weights are (very) sparse, we just write them as a CSV file. This is also what ``MCView`` expect.
    """
    with open(path, "w", encoding="utf8") as file:
        file.write("query,atlas,weight\n")
        for query_index, atlas_index in zip(*weights.nonzero()):
            weight = weights[query_index, atlas_index]
            file.write(f"{qdata.obs_names[query_index]},{adata.obs_names[atlas_index]},{weight}\n")
